from ._data import StochasticSampleMapping
from ._iterables import (
    Kijima1ProcessIterable,
    Kijima2ProcessIterable,
    NonHomogeneousPoissonProcessIterable,
    RenewalProcessIterable,
)

__all__ = [
    "StochasticSampleMapping",
    "NonHomogeneousPoissonProcessIterable",
    "RenewalProcessIterable",
    "Kijima1ProcessIterable",
    "Kijima2ProcessIterable",
]



# pyright: basic
from abc import ABC, abstractmethod
from typing import Generic, Literal, TypeVar, overload

import numpy as np
from numpy.typing import NDArray

from relife.economic import ExponentialDiscounting, Reward
from relife.typing import (
    AnyParametricLifetimeModel,
    NumpyFloat,
)

__all__ = ["OneCycleExpectedCosts", "ReplacementPolicy"]


def _make_timeline(tf: float, nb_steps: int) -> NDArray[np.float64]:
    timeline = np.linspace(0, tf, nb_steps, dtype=np.float64)  # (nb_steps,)
    return np.atleast_2d(timeline)  # (1, nb_steps) to ensure broadcasting


class ExpectedCostsABC(ABC):
    @abstractmethod
    def expected_net_present_value(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""
        The expected net present value.

        .. math::

            z(t) = \mathbb{E}(Z_t) = \int_{0}^{\infty}\mathbb{E}(Z_t~|~X_1 = x)dF(x)

        where :

        - :math:`t` is the time
        - :math:`X_1 \sim F` is the random lifetime of the first asset
        - :math:`Z_t` are the random costs at each time :math:`t`
        - :math:`\delta` is the discounting rate

        It is computed by solving the renewal equation.

        Parameters
        ----------
        tf : float
            The final time.
        nb_steps : int
            The number of steps used to discretized the time.
        total_sum : bool, default False
            If True, returns the total sum over the first axis of the result. If the policy data encodes several
            assets, this option allows to return the sum result on the fleet rather than calling ``np.sum`` afterwards.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline and the computed values.
        """

    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    @abstractmethod
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        r"""
        The asymtotic expected net present value.

        .. math::

            \lim_{t\to\infty} z(t)

        Parameters
        ----------
        total_sum : bool, default False
            If True, returns the total sum over the first axis of the result. If the policy data encodes several
            assets, this option allows to return the sum result on the fleet rather than calling ``np.sum`` afterwards.

        Returns
        -------
        ndarray
            The asymptotic expected values.
        """

    @abstractmethod
    def expected_equivalent_annual_cost(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""
        The expected equivalent annual cost.

        .. math::

            q(t) = \dfrac{\delta z(t)}{1 - e^{-\delta t}}

        where :

        - :math:`t` is the time.
        - :math:`z(t)` is the expected net present value at time :math:`t`.
        - :math:`\delta` is the discounting rate.

        Parameters
        ----------
        tf : float
            The final time.
        nb_steps : int
            The number of steps used to discretized the time.
        total_sum : bool, default False
            If True, returns the total sum over the first axis of the result. If the policy data encodes several
            assets, this option allows to return the sum result on the fleet rather than calling ``np.sum`` afterwards.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline and the computed values.
        """

    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    @abstractmethod
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        r"""
        The asymtotic expected equivalent annual cost.

        .. math::

            \lim_{t\to\infty} q(t)

        Parameters
        ----------
        total_sum : bool, default False
            If True, returns the total sum over the first axis of the result. If the policy data encodes several
            assets, this option allows to return the sum result on the fleet rather than calling ``np.sum`` afterwards.

        Returns
        -------
        ndarray
            The asymptotic expected values.
        """


class OneCycleExpectedCosts(ExpectedCostsABC):
    lifetime_model: AnyParametricLifetimeModel[()]
    reward: Reward
    discounting: ExponentialDiscounting
    period_before_discounting: float

    def __init__(
        self,
        lifetime_model: AnyParametricLifetimeModel[()],
        reward: Reward,
        discounting_rate: float = 0.0,
        period_before_discounting: float = 1.0,
    ) -> None:
        self.reward = reward
        self.discounting = ExponentialDiscounting(discounting_rate)
        if period_before_discounting <= 0:
            raise ValueError("The period_before_discounting must be greater than 0")
        self.period_before_discounting = period_before_discounting
        self.lifetime_model = lifetime_model

    def expected_net_present_value(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        timeline = _make_timeline(tf, nb_steps)
        etc = np.asarray(
            self.lifetime_model.ls_integrate(
                lambda x: self.reward.conditional_expectation(x)
                * self.discounting.factor(x),
                np.zeros_like(timeline),
                timeline,
                deg=15,
            ),
            dtype=float,
        )  # (nb_steps,) or (m, nb_steps)
        if total_sum and etc.ndim == 2:
            etc = np.sum(etc, axis=0)
        if timeline.ndim == 2:
            timeline = timeline[0, :]
        return timeline, etc  # (nb_steps,) and (nb_steps,)/(m, nb_steps)

    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        # reward partial expectation
        value = np.squeeze(
            self.lifetime_model.ls_integrate(
                lambda x: self.reward.conditional_expectation(x)
                * self.discounting.factor(x),
                np.float64(0.0),
                np.asarray(np.inf),
                deg=15,
            )
        )
        if total_sum:
            value = np.sum(value)
        return value  # () or (m,)

    def _expected_equivalent_annual_cost(
        self, timeline: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        # timeline : (nb_steps,) or (m, nb_steps)
        def f(x: NDArray[np.float64]) -> NDArray[np.float64]:
            # avoid zero division + 1e-6
            return (
                self.reward.conditional_expectation(x)
                * self.discounting.factor(x)
                / (self.discounting.annuity_factor(x) + 1e-6)
            )

        q0 = self.lifetime_model.cdf(self.period_before_discounting) * f(
            np.asarray(self.period_before_discounting, dtype=float)
        )  # () or (m, 1)
        a = np.full_like(
            timeline, self.period_before_discounting
        )  # (nb_steps,) or (m, nb_steps)

        # change first value of lower bound to compute the integral
        a[timeline < self.period_before_discounting] = 0.0  # (nb_steps,)
        # a = np.where(timeline < self.period_before_discounting, 0., a)  # (nb_steps,)
        integral = self.lifetime_model.ls_integrate(
            f, a, timeline, deg=15
        )  # (nb_steps,) or (m, nb_steps) if q0: (), or (m, nb_steps) if q0 : (m, 1)
        mask = np.broadcast_to(
            timeline < self.period_before_discounting, integral.shape
        )  # (), (nb_steps,) or (m, nb_steps)
        q0 = np.broadcast_to(q0, integral.shape)  # (nb_steps,) or (m, nb_steps)
        integral = np.where(mask, q0, q0 + integral)
        return np.squeeze(integral)  # (nb_steps,)/(m, nb_steps)

    def expected_equivalent_annual_cost(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        timeline = _make_timeline(tf, nb_steps)  # (nb_steps,) or (m, nb_steps)
        value = self._expected_equivalent_annual_cost(timeline)
        if timeline.ndim == 2:
            timeline = timeline[0, :]  # (nb_steps,)
        if total_sum and value.ndim == 2:
            value = np.sum(value, axis=0)
        return timeline, value  # (nb_steps,) or (m, nb_steps)

    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        timeline = np.atleast_2d(np.array(np.inf))  # (1, 1) to ensure broadcasting
        value = self._expected_equivalent_annual_cost(timeline)
        if total_sum:
            value = np.sum(value)
        return value  # () or (m,)


M = TypeVar("M")


class ReplacementPolicy(ExpectedCostsABC, Generic[M], ABC):
    baseline_model: M
    discounting_rate: float
    _cost_structure: dict[str, NumpyFloat]

    def __init__(
        self,
        baseline_model: M,
        cost_structure: dict[str, NumpyFloat],
        discounting_rate: float = 0.0,
    ):
        self.baseline_model = baseline_model
        self.discounting_rate = discounting_rate
        self._cost_structure = cost_structure  # hidden, contains reshaped cost arrays



# pyright: basic

import numpy as np
from pytest import approx
from scipy.stats import boxcox, zscore

from relife.lifetime_model import (
    ParametricAcceleratedFailureTime,
    ParametricProportionalHazard,
    Weibull,
)
from relife.lifetime_model._regression import LinearCovarEffect


def expected_shape(**kwargs):
    def shape_contrib(**_kwargs):
        yield ()  # yield at least (), in case kwargs is empty
        for k, v in _kwargs.items():
            match k:
                case "covar" if v.ndim == 2:
                    yield v.shape[0], 1
                case "covar" if v.ndim < 2:
                    yield ()
                case _:
                    yield v.shape

    return np.broadcast_shapes(*tuple(shape_contrib(**kwargs)))


def rvs_expected_shape(size, **kwargs):
    out_shape = expected_shape(**kwargs)
    if size != 1:
        return np.broadcast_shapes(out_shape, (size,))
    return out_shape


def test_covar_effect():
    """
    covar : () or (nb_coef,)
    => g : ()
    => jac_g : (nb_coef,)

    covar : (m, nb_coef)
    => g : (m, 1)
    => jac_g : (nb_coef, m, 1)
    """

    covar_effect = LinearCovarEffect(coefficients=(2.4, 5.5))
    z1 = np.array([1, 2, 3])
    z2 = np.array([0.8, 0.7, 0.5])
    assert covar_effect.g(np.column_stack((z1, z2))) == approx(
        np.exp(2.4 * z1 + 5.5 * z2).reshape(-1, 1)
    )
    assert covar_effect.jac_g(np.column_stack((z1, z2)))[0] == approx(
        (z1 * np.exp(2.4 * z1 + 5.5 * z2)).reshape(-1, 1)
    )
    assert covar_effect.jac_g(np.column_stack((z1, z2)))[1] == approx(
        (z2 * np.exp(2.4 * z1 + 5.5 * z2)).reshape(-1, 1)
    )

    assert covar_effect.g(np.ones(covar_effect.nb_coef)).shape == ()
    assert covar_effect.g(np.ones((1, covar_effect.nb_coef))).shape == (1, 1)
    assert covar_effect.g(np.ones((10, covar_effect.nb_coef))).shape == (10, 1)

    assert covar_effect.jac_g(np.ones(covar_effect.nb_coef)).shape == (
        covar_effect.nb_coef,
    )
    assert covar_effect.jac_g(np.ones(covar_effect.nb_coef)).shape == (
        covar_effect.nb_coef,
    )
    assert covar_effect.jac_g(np.ones((1, covar_effect.nb_coef))).shape == (
        covar_effect.nb_coef,
        1,
        1,
    )
    assert covar_effect.jac_g(np.ones((10, covar_effect.nb_coef))).shape == (
        covar_effect.nb_coef,
        10,
        1,
    )


def test_rvs(regression, covar, rvs_size):
    assert regression.rvs(rvs_size, covar).shape == rvs_expected_shape(
        rvs_size, covar=covar
    )
    assert all(
        arr.shape == rvs_expected_shape(rvs_size, covar=covar)
        for arr in regression.rvs(rvs_size, covar, return_event=True)
    )
    assert all(
        arr.shape == rvs_expected_shape(rvs_size, covar=covar)
        for arr in regression.rvs(rvs_size, covar, return_entry=True)
    )
    assert all(
        arr.shape == rvs_expected_shape(rvs_size, covar=covar)
        for arr in regression.rvs(
            rvs_size,
            covar,
            return_event=True,
            return_entry=True,
        )
    )


def test_sf(regression, time, covar):
    assert regression.sf(time, covar).shape == expected_shape(time=time, covar=covar)


def test_hf(regression, time, covar):
    assert regression.hf(time, covar).shape == expected_shape(time=time, covar=covar)


def test_chf(regression, time, covar):
    assert regression.chf(time, covar).shape == expected_shape(time=time, covar=covar)


def test_cdf(regression, time, covar):
    assert regression.cdf(time, covar).shape == expected_shape(time=time, covar=covar)


def test_pdf(regression, time, covar):
    assert regression.pdf(time, covar).shape == expected_shape(time=time, covar=covar)


def test_ppf(regression, probability, covar):
    assert regression.ppf(probability, covar).shape == expected_shape(
        time=probability, covar=covar
    )


def test_ichf(regression, probability, covar):
    assert regression.ichf(probability, covar).shape == expected_shape(
        time=probability, covar=covar
    )


def test_isf(regression, probability, covar):
    assert regression.isf(probability, covar).shape == expected_shape(
        time=probability, covar=covar
    )
    assert regression.isf(np.full(probability.shape, 0.5), covar) == approx(
        np.broadcast_to(
            regression.median(covar), expected_shape(time=probability, covar=covar)
        )
    )


def test_dhf(regression, time, covar):
    assert regression.dhf(time, covar).shape == expected_shape(time=time, covar=covar)


def test_jac_sf(regression, time, covar):
    assert regression.jac_sf(time, covar).shape == (
        regression.nb_params,
    ) + expected_shape(time=time, covar=covar)


def test_jac_hf(regression, time, covar):
    assert regression.jac_hf(time, covar).shape == (
        regression.nb_params,
    ) + expected_shape(time=time, covar=covar)


def test_jac_chf(regression, time, covar):
    assert regression.jac_chf(time, covar).shape == (
        regression.nb_params,
    ) + expected_shape(time=time, covar=covar)


def test_jac_cdf(regression, time, covar):
    assert regression.jac_cdf(time, covar).shape == (
        regression.nb_params,
    ) + expected_shape(time=time, covar=covar)


def test_jac_pdf(regression, time, covar):
    assert regression.jac_pdf(time, covar).shape == (
        regression.nb_params,
    ) + expected_shape(time=time, covar=covar)


def test_ls_integrate(regression, integration_bound_a, integration_bound_b, covar):
    # integral_a^b dF(x)
    integration = regression.ls_integrate(
        np.ones_like, integration_bound_a, integration_bound_b, covar, deg=100
    )
    assert integration.shape == expected_shape(
        a=integration_bound_a, b=integration_bound_b, covar=covar
    )
    assert integration == approx(
        regression.cdf(integration_bound_b, covar)
        - regression.cdf(integration_bound_a, covar)
    )
    # integral_0^inf x*dF(x)
    integration = regression.ls_integrate(
        lambda x: x,
        np.zeros_like(integration_bound_a),
        np.full_like(integration_bound_b, np.inf),
        covar,
        deg=100,
    )
    assert integration == approx(
        np.broadcast_to(
            regression.mean(covar),
            expected_shape(a=integration_bound_a, b=integration_bound_b, covar=covar),
        ),
        rel=1e-3,
    )


def test_aft_pph_weibull_eq(insulator_string_data):
    covar_data = zscore(
        np.column_stack(
            (
                boxcox(insulator_string_data["pHCl"])[0],
                boxcox(insulator_string_data["pH2SO4"])[0],
                boxcox(insulator_string_data["HNO3"])[0],
            )
        )
    )
    weibull_aft = ParametricAcceleratedFailureTime(Weibull()).fit(
        insulator_string_data["time"],
        covar_data,
        event=insulator_string_data["event"],
        entry=insulator_string_data["entry"],
    )
    weibull_pph = ParametricProportionalHazard(Weibull()).fit(
        insulator_string_data["time"],
        covar_data,
        event=insulator_string_data["event"],
        entry=insulator_string_data["entry"],
    )

    assert weibull_pph.baseline.params == approx(weibull_aft.baseline.params, rel=1e-3)
    assert weibull_pph.covar_effect.params == approx(
        -weibull_aft.baseline.params[0] * weibull_aft.covar_effect.params,
        rel=1e-3,
    )


def test_negative_log(regression_likelihood):
    params = regression_likelihood.model.params.copy()
    assert isinstance(regression_likelihood.negative_log(params), float)


def test_jac_negative_log(regression_likelihood):
    params = regression_likelihood.model.params.copy()
    assert regression_likelihood.jac_negative_log(params).shape == (params.size,)



# pyright: basic

import numpy as np

from ._array_utils import get_args_nb_assets, reshape_1d_arg

__all__ = [
    "get_model_nb_assets",
]


def get_model_nb_assets(model):
    """
    Gets the number of assets stored by a model (frozen or not).
    """
    from relife.base import FrozenParametricModel
    from relife.lifetime_model import EquilibriumDistribution, MinimumDistribution
    from relife.lifetime_model._regression import ParametricLifetimeRegression
    from relife.stochastic_process import NonHomogeneousPoissonProcess, RenewalProcess

    if isinstance(model, EquilibriumDistribution) or isinstance(
        model, MinimumDistribution
    ):
        return get_model_nb_assets(model.baseline)

    if isinstance(model, NonHomogeneousPoissonProcess):
        return get_model_nb_assets(model.lifetime_model)

    if isinstance(model, RenewalProcess):
        lifetime_model_nb_assets = get_model_nb_assets(model.lifetime_model)
        if model.first_lifetime_model is not None:
            first_lifetime_model_nb_assets = get_model_nb_assets(
                model.first_lifetime_model
            )
            return max(lifetime_model_nb_assets, first_lifetime_model_nb_assets)
        return lifetime_model_nb_assets

    if isinstance(model, FrozenParametricModel):
        if isinstance(model._unfrozen_model, NonHomogeneousPoissonProcess):
            return get_model_nb_assets(model._unfrozen_model)
        if isinstance(model._unfrozen_model, ParametricLifetimeRegression):
            # specific covar reshape
            reshaped_args = [np.atleast_2d(model.args[0])]
            reshaped_args += [reshape_1d_arg(arg) for arg in model.args[1:]]
        else:
            reshaped_args = [reshape_1d_arg(arg) for arg in model.args]
        return get_args_nb_assets(*reshaped_args)

    return 1



# pyright: basic

import numpy as np
from pytest import approx

from relife.lifetime_model import AgeReplacementModel, LeftTruncatedModel


def expected_shape(**kwargs):
    def shape_contrib(**_kwargs):
        yield ()  # yield at least (), in case kwargs is empty
        for k, v in _kwargs.items():
            match k:
                case "covar" if v.ndim == 2:
                    yield v.shape[0], 1
                case "covar" if v.ndim < 2:
                    yield ()
                case "ar" | "a0" if v.ndim == 2 or v.ndim == 0:
                    yield v.shape
                case "ar" | "a0" if v.ndim == 1:
                    yield v.size, 1
                case _:
                    yield v.shape

    return np.broadcast_shapes(*tuple(shape_contrib(**kwargs)))


def rvs_expected_shape(size, **kwargs):
    out_shape = expected_shape(**kwargs)
    if size != 1:
        return np.broadcast_shapes(out_shape, (size,))
    return out_shape


class TestAgeReplacementDistribution:
    def test_rvs(self, distribution, ar, rvs_size):
        assert AgeReplacementModel(distribution).rvs(
            rvs_size, ar
        ).shape == rvs_expected_shape(rvs_size, ar=ar)
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, ar=ar)
            for arr in AgeReplacementModel(distribution).rvs(
                rvs_size, ar, return_event=True
            )
        )
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, ar=ar)
            for arr in AgeReplacementModel(distribution).rvs(
                rvs_size, ar, return_entry=True
            )
        )
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, ar=ar)
            for arr in AgeReplacementModel(distribution).rvs(
                rvs_size,
                ar,
                return_event=True,
                return_entry=True,
            )
        )

    def test_sf(self, distribution, time, ar):
        assert AgeReplacementModel(distribution).sf(time, ar).shape == expected_shape(
            time=time, ar=ar
        )

    def test_hf(self, distribution, time, ar):
        assert AgeReplacementModel(distribution).hf(time, ar).shape == expected_shape(
            time=time, ar=ar
        )

    def test_chf(self, distribution, time, ar):
        assert AgeReplacementModel(distribution).chf(time, ar).shape == expected_shape(
            time=time, ar=ar
        )

    def test_cdf(self, distribution, time, ar):
        assert AgeReplacementModel(distribution).cdf(time, ar).shape == expected_shape(
            time=time, ar=ar
        )

    def test_pdf(self, distribution, time, ar):
        assert AgeReplacementModel(distribution).pdf(time, ar).shape == expected_shape(
            time=time, ar=ar
        )

    def test_ppf(self, distribution, time, ar):
        assert AgeReplacementModel(distribution).ppf(time, ar).shape == expected_shape(
            time=time, ar=ar
        )

    def test_ichf(self, distribution, probability, ar):
        assert AgeReplacementModel(distribution).ichf(
            probability, ar
        ).shape == expected_shape(probability=probability, ar=ar)

    def test_isf(self, distribution, probability, ar):
        out_shape = expected_shape(probability=probability, ar=ar)
        assert AgeReplacementModel(distribution).isf(probability, ar).shape == out_shape
        assert AgeReplacementModel(distribution).isf(
            np.full(out_shape, 0.5), ar
        ) == approx(
            np.broadcast_to(AgeReplacementModel(distribution).median(ar), out_shape)
        )

    def test_moment(self, distribution, ar):
        assert AgeReplacementModel(distribution).moment(1, ar).shape == expected_shape(
            ar=ar
        )
        assert AgeReplacementModel(distribution).moment(2, ar).shape == expected_shape(
            ar=ar
        )

    def test_mean(self, distribution, ar):
        assert AgeReplacementModel(distribution).mean(ar).shape == expected_shape(ar=ar)

    def test_var(self, distribution, ar):
        assert AgeReplacementModel(distribution).var(ar).shape == expected_shape(ar=ar)

    def test_median(self, distribution, ar):
        assert AgeReplacementModel(distribution).median(ar).shape == expected_shape(
            ar=ar
        )

    def test_ls_integrate(
        self, distribution, integration_bound_a, integration_bound_b, ar
    ):
        np.random.seed(10)
        ar = np.random.uniform(2.0, 8.0, size=ar.shape)
        # integral_a^b dF(x)
        out_shape = expected_shape(
            integration_bound_a=integration_bound_a,
            integration_bound_b=integration_bound_b,
            ar=ar,
        )
        integration = AgeReplacementModel(distribution).ls_integrate(
            np.ones_like, integration_bound_a, integration_bound_b, ar, deg=100
        )
        assert integration.shape == out_shape
        assert integration == approx(
            AgeReplacementModel(distribution).cdf(integration_bound_b, ar)
            - AgeReplacementModel(distribution).cdf(integration_bound_a, ar)
        )
        # integral_0^inf x*dF(x)
        integration = AgeReplacementModel(distribution).ls_integrate(
            lambda x: x,
            np.zeros_like(integration_bound_a),
            np.full_like(integration_bound_b, np.inf),
            ar,
            deg=100,
        )
        assert integration == approx(
            np.broadcast_to(AgeReplacementModel(distribution).mean(ar), out_shape),
            rel=1e-3,
        )


class TestAgeReplacementRegression:
    def test_rvs(self, regression, ar, covar, rvs_size):
        assert AgeReplacementModel(regression).rvs(
            rvs_size,
            ar,
            covar,
        ).shape == rvs_expected_shape(rvs_size, ar=ar, covar=covar)
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, ar=ar, covar=covar)
            for arr in AgeReplacementModel(regression).rvs(
                rvs_size, ar, covar, return_event=True
            )
        )
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, ar=ar, covar=covar)
            for arr in AgeReplacementModel(regression).rvs(
                rvs_size, ar, covar, return_entry=True
            )
        )
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, ar=ar, covar=covar)
            for arr in AgeReplacementModel(regression).rvs(
                rvs_size,
                ar,
                covar,
                return_event=True,
                return_entry=True,
            )
        )

    def test_sf(self, regression, time, ar, covar):
        assert AgeReplacementModel(regression).sf(
            time, ar, covar
        ).shape == expected_shape(time=time, ar=ar, covar=covar)

    def test_hf(self, regression, time, ar, covar):
        assert AgeReplacementModel(regression).hf(
            time, ar, covar
        ).shape == expected_shape(time=time, ar=ar, covar=covar)

    def test_chf(self, regression, time, ar, covar):
        assert AgeReplacementModel(regression).chf(
            time, ar, covar
        ).shape == expected_shape(time=time, ar=ar, covar=covar)

    def test_cdf(self, regression, time, ar, covar):
        assert AgeReplacementModel(regression).cdf(
            time, ar, covar
        ).shape == expected_shape(time=time, ar=ar, covar=covar)

    def test_pdf(self, regression, time, ar, covar):
        assert AgeReplacementModel(regression).pdf(
            time, ar, covar
        ).shape == expected_shape(time=time, ar=ar, covar=covar)

    def test_ppf(self, regression, time, ar, covar):
        assert AgeReplacementModel(regression).ppf(
            time, ar, covar
        ).shape == expected_shape(time=time, ar=ar, covar=covar)

    def test_ichf(self, regression, probability, ar, covar):
        assert AgeReplacementModel(regression).ichf(
            probability, ar, covar
        ).shape == expected_shape(probability=probability, ar=ar, covar=covar)

    def test_isf(self, regression, probability, ar, covar):
        out_shape = expected_shape(probability=probability, ar=ar, covar=covar)
        assert (
            AgeReplacementModel(regression).isf(probability, ar, covar).shape
            == out_shape
        )
        assert AgeReplacementModel(regression).isf(
            np.full(out_shape, 0.5), ar, covar
        ) == approx(
            np.broadcast_to(
                AgeReplacementModel(regression).median(ar, covar), out_shape
            )
        )

    def test_moment(self, regression, ar, covar):
        assert AgeReplacementModel(regression).moment(
            1, ar, covar
        ).shape == expected_shape(ar=ar, covar=covar)
        assert AgeReplacementModel(regression).moment(
            2, ar, covar
        ).shape == expected_shape(ar=ar, covar=covar)

    def test_mean(self, regression, ar, covar):
        assert AgeReplacementModel(regression).mean(ar, covar).shape == expected_shape(
            ar=ar, covar=covar
        )

    def test_var(self, regression, ar, covar):
        assert AgeReplacementModel(regression).var(ar, covar).shape == expected_shape(
            ar=ar, covar=covar
        )

    def test_median(self, regression, ar, covar):
        assert AgeReplacementModel(regression).median(
            ar, covar
        ).shape == expected_shape(ar=ar, covar=covar)

    def test_ls_integrate(
        self, regression, integration_bound_a, integration_bound_b, ar, covar
    ):
        np.random.seed(10)
        ar = np.random.uniform(2.0, 8.0, size=ar.shape)
        # integral_a^b dF(x)
        out_shape = expected_shape(
            integration_bound_a=integration_bound_a,
            integration_bound_b=integration_bound_b,
            ar=ar,
            covar=covar,
        )
        integration = AgeReplacementModel(regression).ls_integrate(
            np.ones_like, integration_bound_a, integration_bound_b, ar, covar, deg=100
        )
        assert integration.shape == out_shape
        assert integration == approx(
            AgeReplacementModel(regression).cdf(integration_bound_b, ar, covar)
            - AgeReplacementModel(regression).cdf(integration_bound_a, ar, covar)
        )
        # integral_0^inf x*dF(x)
        integration = AgeReplacementModel(regression).ls_integrate(
            lambda x: x,
            np.zeros_like(integration_bound_a),
            np.full_like(integration_bound_b, np.inf),
            ar,
            covar,
            deg=100,
        )
        assert integration == approx(
            np.broadcast_to(AgeReplacementModel(regression).mean(ar, covar), out_shape),
            rel=1e-3,
        )


class TestLeftTruncatedDistribution:
    def test_rvs(self, distribution, a0, rvs_size):
        assert LeftTruncatedModel(distribution).rvs(
            rvs_size,
            a0,
        ).shape == rvs_expected_shape(rvs_size, a0=a0)
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, a0=a0)
            for arr in LeftTruncatedModel(distribution).rvs(
                rvs_size, a0, return_event=True
            )
        )
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, a0=a0)
            for arr in LeftTruncatedModel(distribution).rvs(
                rvs_size, a0, return_entry=True
            )
        )
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, a0=a0)
            for arr in LeftTruncatedModel(distribution).rvs(
                rvs_size,
                a0,
                return_event=True,
                return_entry=True,
            )
        )

    def test_sf(self, distribution, time, a0):
        assert LeftTruncatedModel(distribution).sf(time, a0).shape == expected_shape(
            time=time, a0=a0
        )

    def test_hf(self, distribution, time, a0):
        assert LeftTruncatedModel(distribution).hf(time, a0).shape == expected_shape(
            time=time, a0=a0
        )

    def test_chf(self, distribution, time, a0):
        assert LeftTruncatedModel(distribution).chf(time, a0).shape == expected_shape(
            time=time, a0=a0
        )

    def test_cdf(self, distribution, time, a0):
        assert LeftTruncatedModel(distribution).cdf(time, a0).shape == expected_shape(
            time=time, a0=a0
        )

    def test_pdf(self, distribution, time, a0):
        assert LeftTruncatedModel(distribution).pdf(time, a0).shape == expected_shape(
            time=time, a0=a0
        )

    def test_ppf(self, distribution, time, a0):
        assert LeftTruncatedModel(distribution).ppf(time, a0).shape == expected_shape(
            time=time, a0=a0
        )

    def test_ichf(self, distribution, probability, a0):
        assert LeftTruncatedModel(distribution).ichf(
            probability, a0
        ).shape == expected_shape(probability=probability, a0=a0)

    def test_isf(self, distribution, probability, a0):
        out_shape = expected_shape(probability=probability, a0=a0)
        assert LeftTruncatedModel(distribution).isf(probability, a0).shape == out_shape
        assert LeftTruncatedModel(distribution).isf(
            np.full(out_shape, 0.5), a0
        ) == approx(
            np.broadcast_to(LeftTruncatedModel(distribution).median(a0), out_shape)
        )

    def test_moment(self, distribution, a0):
        assert LeftTruncatedModel(distribution).moment(1, a0).shape == expected_shape(
            a0=a0
        )
        assert LeftTruncatedModel(distribution).moment(2, a0).shape == expected_shape(
            a0=a0
        )

    def test_mean(self, distribution, a0):
        assert LeftTruncatedModel(distribution).mean(a0).shape == expected_shape(a0=a0)

    def test_var(self, distribution, a0):
        assert LeftTruncatedModel(distribution).var(a0).shape == expected_shape(a0=a0)

    def test_median(self, distribution, a0):
        assert LeftTruncatedModel(distribution).median(a0).shape == expected_shape(
            a0=a0
        )

    def test_ls_integrate(
        self, distribution, integration_bound_a, integration_bound_b, a0
    ):
        # integral_a^b dF(x)
        out_shape = expected_shape(
            integration_bound_a=integration_bound_a,
            integration_bound_b=integration_bound_b,
            a0=a0,
        )
        integration = LeftTruncatedModel(distribution).ls_integrate(
            np.ones_like, integration_bound_a, integration_bound_b, a0, deg=100
        )
        assert integration.shape == out_shape
        assert integration == approx(
            LeftTruncatedModel(distribution).cdf(integration_bound_b, a0)
            - LeftTruncatedModel(distribution).cdf(integration_bound_a, a0)
        )
        # integral_0^inf x*dF(x)
        integration = LeftTruncatedModel(distribution).ls_integrate(
            lambda x: x,
            np.zeros_like(integration_bound_a),
            np.full_like(integration_bound_b, np.inf),
            a0,
            deg=100,
        )
        assert integration == approx(
            np.broadcast_to(LeftTruncatedModel(distribution).mean(a0), out_shape),
            rel=1e-3,
        )


class TestLeftTruncatedRegression:
    def test_rvs(self, regression, a0, covar, rvs_size):
        assert LeftTruncatedModel(regression).rvs(
            rvs_size, a0, covar
        ).shape == rvs_expected_shape(rvs_size, a0=a0, covar=covar)
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, a0=a0, covar=covar)
            for arr in LeftTruncatedModel(regression).rvs(
                rvs_size, a0, covar, return_event=True
            )
        )
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, a0=a0, covar=covar)
            for arr in LeftTruncatedModel(regression).rvs(
                rvs_size, a0, covar, return_entry=True
            )
        )
        assert all(
            arr.shape == rvs_expected_shape(rvs_size, a0=a0, covar=covar)
            for arr in LeftTruncatedModel(regression).rvs(
                rvs_size,
                a0,
                covar,
                return_event=True,
                return_entry=True,
            )
        )

    def test_sf(self, regression, time, a0, covar):
        assert LeftTruncatedModel(regression).sf(
            time, a0, covar
        ).shape == expected_shape(time=time, a0=a0, covar=covar)

    def test_hf(self, regression, time, a0, covar):
        assert LeftTruncatedModel(regression).hf(
            time, a0, covar
        ).shape == expected_shape(time=time, a0=a0, covar=covar)

    def test_chf(self, regression, time, a0, covar):
        assert LeftTruncatedModel(regression).chf(
            time, a0, covar
        ).shape == expected_shape(time=time, a0=a0, covar=covar)

    def test_cdf(self, regression, time, a0, covar):
        assert LeftTruncatedModel(regression).cdf(
            time, a0, covar
        ).shape == expected_shape(time=time, a0=a0, covar=covar)

    def test_pdf(self, regression, time, a0, covar):
        assert LeftTruncatedModel(regression).pdf(
            time, a0, covar
        ).shape == expected_shape(time=time, a0=a0, covar=covar)

    def test_ppf(self, regression, time, a0, covar):
        assert LeftTruncatedModel(regression).ppf(
            time, a0, covar
        ).shape == expected_shape(time=time, a0=a0, covar=covar)

    def test_ichf(self, regression, probability, a0, covar):
        assert LeftTruncatedModel(regression).ichf(
            probability, a0, covar
        ).shape == expected_shape(probability=probability, a0=a0, covar=covar)

    def test_isf(self, regression, probability, a0, covar):
        out_shape = expected_shape(probability=probability, a0=a0, covar=covar)
        assert (
            LeftTruncatedModel(regression).isf(probability, a0, covar).shape
            == out_shape
        )
        assert LeftTruncatedModel(regression).isf(
            np.full(out_shape, 0.5), a0, covar
        ) == approx(
            np.broadcast_to(LeftTruncatedModel(regression).median(a0, covar), out_shape)
        )

    def test_moment(self, regression, a0, covar):
        assert LeftTruncatedModel(regression).moment(
            1, a0, covar
        ).shape == expected_shape(a0=a0, covar=covar)
        assert LeftTruncatedModel(regression).moment(
            2, a0, covar
        ).shape == expected_shape(a0=a0, covar=covar)

    def test_mean(self, regression, a0, covar):
        assert LeftTruncatedModel(regression).mean(a0, covar).shape == expected_shape(
            a0=a0, covar=covar
        )

    def test_var(self, regression, a0, covar):
        assert LeftTruncatedModel(regression).var(a0, covar).shape == expected_shape(
            a0=a0, covar=covar
        )

    def test_median(self, regression, a0, covar):
        assert LeftTruncatedModel(regression).median(a0, covar).shape == expected_shape(
            a0=a0, covar=covar
        )

    def test_ls_integrate(
        self, regression, integration_bound_a, integration_bound_b, a0, covar
    ):
        # integral_a^b dF(x)
        out_shape = expected_shape(
            integration_bound_a=integration_bound_a,
            integration_bound_b=integration_bound_b,
            a0=a0,
            covar=covar,
        )
        np.random.seed(10)
        a0 = np.random.uniform(2.0, 8.0, size=a0.shape)
        integration = LeftTruncatedModel(regression).ls_integrate(
            np.ones_like, integration_bound_a, integration_bound_b, a0, covar, deg=100
        )
        assert integration.shape == out_shape
        assert integration == approx(
            LeftTruncatedModel(regression).cdf(integration_bound_b, a0, covar)
            - LeftTruncatedModel(regression).cdf(integration_bound_a, a0, covar)
        )
        # integral_0^inf x*dF(x)
        integration = LeftTruncatedModel(regression).ls_integrate(
            lambda x: x,
            np.zeros_like(integration_bound_a),
            np.full_like(integration_bound_b, np.inf),
            a0,
            covar,
            deg=100,
        )
        assert integration == approx(
            np.broadcast_to(LeftTruncatedModel(regression).mean(a0, covar), out_shape),
            rel=1e-3,
        )



"""
``relife.lifetime_model``
=========================

The ReLife lifetime_model module exposes various stochastic models to modelize
lifetime data. Internal operations are computed using NumPy and Scipy.

- NumPy: https://github.com/numpy/numpy
- Scipy: https://github.com/scipy/scipy

Objects present in relife.lifetime_model are listed below.

Lifetime distribution
---------------------

    Exponential
    Weibull
    Gompertz
    Gamma
    LogLogistic
    MinimumDistribution
    EquilibriumDistribution


Lifetime regression
-------------------

    ParametricProportionalHazard
    ParametricAcceleratedFailureTime

Semiparametric lifetime regression
----------------------------------

    SemiParametricProportionalHazard


Nonparametric models
--------------------

    KaplanMeier
    ECDF
    NelsonAalen

Conditional models
------------------

    AgeReplacementModel
    LeftTruncatedModel


Likelihoods
-----------

    LifetimeLikelihood
    CoxPartialLifetimeLikelihood
    BreslowPartialLifetimeLikelihood
    EfronPartialLifetimeLikelihood

"""

from ._base import LifetimeLikelihood
from ._conditional_model import AgeReplacementModel, LeftTruncatedModel
from ._distribution import (
    EquilibriumDistribution,
    Exponential,
    Gamma,
    Gompertz,
    LogLogistic,
    MinimumDistribution,
    Weibull,
)
from ._non_parametric import ECDF, KaplanMeier, NelsonAalen
from ._regression import ParametricAcceleratedFailureTime, ParametricProportionalHazard
from ._semi_parametric import (
    BreslowPartialLifetimeLikelihood,
    CoxPartialLifetimeLikelihood,
    EfronPartialLifetimeLikelihood,
    SemiParametricProportionalHazard,
)

__all__ = [
    "LifetimeLikelihood",
    "Exponential",
    "Weibull",
    "Gompertz",
    "Gamma",
    "LogLogistic",
    "MinimumDistribution",
    "EquilibriumDistribution",
    "ParametricProportionalHazard",
    "ParametricAcceleratedFailureTime",
    "SemiParametricProportionalHazard",
    "KaplanMeier",
    "ECDF",
    "NelsonAalen",
    "AgeReplacementModel",
    "LeftTruncatedModel",
    "CoxPartialLifetimeLikelihood",
    "EfronPartialLifetimeLikelihood",
    "BreslowPartialLifetimeLikelihood",
]



from typing import Callable, overload

import numpy as np
from numpy.typing import NDArray

from relife.typing import AnyFloat, NumpyFloat

__all__ = [
    "legendre_quadrature",
    "laguerre_quadrature",
    "unweighted_laguerre_quadrature",
    "broadcast_bounds",
]


def _control_shape(bound: AnyFloat) -> NDArray[np.float64]:
    arr = np.asarray(bound, dtype=np.float64)
    if np.any(arr < 0):
        raise ValueError("Bound values of the integral can't be lower than 0")
    if arr.ndim > 2:
        raise ValueError("Bound the integral can't have more than 2 dimensions")
    return arr


@overload
def broadcast_bounds(a: AnyFloat) -> NDArray[np.float64]: ...
@overload
def broadcast_bounds(
    a: AnyFloat, b: AnyFloat
) -> tuple[NDArray[np.float64], NDArray[np.float64]]: ...
def broadcast_bounds(
    a: AnyFloat, b: AnyFloat | None = None
) -> NDArray[np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:
    a = _control_shape(a)
    if b is not None:
        b = _control_shape(b)
        try:
            a, b = np.broadcast_arrays(a, b)
            return a.copy(), b.copy()
        except ValueError as err:
            raise ValueError(
                f"Incompatible a, b shapes. Got a.shape, b.shape : {a.shape}, {b.shape}"
            ) from err
    return a


def legendre_quadrature(
    func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
    a: AnyFloat,
    b: AnyFloat,
    deg: int = 10,
) -> NumpyFloat:
    r"""Numerical integration of :math:`f(x)` over the interval :math:`[a,b]`

    `func` must accept (deg,), (deg, n) or (deg, m, n) array shapes
    a can be zero
    b must not be inf

    a, b shapes can be either 0d (float like), 1d or 2d
    """
    arr_a, arr_b = broadcast_bounds(a, b)  # () or (n,) or (m, n)
    quad = np.polynomial.legendre.leggauss(deg)  # (deg,)
    x, w = quad
    x = np.expand_dims(
        x, axis=tuple(range(1, arr_a.ndim + 1))
    )  # (deg,), (deg, 1) or (deg, 1, 1)
    w = np.expand_dims(
        w, axis=tuple(range(1, arr_a.ndim + 1))
    )  # (deg,), (deg, 1) or (deg, 1, 1)

    if np.any(arr_b == np.inf):
        raise ValueError("Bound values of Legendre quadrature must be finite")
    if np.any(arr_a > arr_b):
        raise ValueError("Bound values a must be lower than values of b")

    p = (arr_b - arr_a) / 2  # () or (n,) or (m, n)
    m = (arr_a + arr_b) / 2  # () or (n,) or (m, n)
    u = p * x + m  # (deg,) or (deg, n) or (deg, m, n)
    v = p * w  # (deg,) or (deg, n) or (deg, m, n)
    fvalues = func(
        u
    )  # (d_1, ..., d_i, deg) or (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
    try:
        _ = np.broadcast_shapes(fvalues.shape[-len(u.shape) :], u.shape)
    except ValueError:
        raise ValueError(
            f"""
            func can't squeeze input dimensions. If x has shape (d_1, ..., d_i), func(x) must have shape (..., d_1, ..., d_i).
            Ex : if x.shape == (m, n), func(x).shape == (..., m, n).
            """
        )

    return np.sum(
        v * fvalues, axis=-v.ndim
    )  # (d_1, ..., d_i) or (d_1, ..., d_i, n) or (d_1, ..., d_i, m, n)


def laguerre_quadrature(
    func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
    a: AnyFloat,
    deg: int = 10,
) -> NumpyFloat:
    r"""Numerical integration of :math:`f(x) * exp(-x)` over the interval :math:`[a, \infty]`

    `func` must accept (deg,), (deg, n) or (deg, m, n) array shapes
    It must handle at least 3 dimensions.
    a can be zero with ndim <= 2.
    """
    arr_a = broadcast_bounds(a)  # () or (n,) or (m, n)
    quad = np.polynomial.laguerre.laggauss(deg)  # (deg,)
    x, w = quad
    x = np.expand_dims(
        x, axis=tuple(range(1, arr_a.ndim + 1))
    )  # (deg,), (deg, 1) or (deg, 1, 1)
    w = np.expand_dims(
        w, axis=tuple(range(1, arr_a.ndim + 1))
    )  # (deg,), (deg, 1) or (deg, 1, 1)

    shifted_x = x + arr_a  # (deg,) or (deg, n) or (deg, m, n)
    fvalues = func(
        shifted_x
    )  # (d_1, ..., d_i, deg) or (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
    try:
        _ = np.broadcast_shapes(fvalues.shape[-len(shifted_x.shape) :], shifted_x.shape)
    except ValueError:
        # func est une fonction réel univariée et pas multivariée
        raise ValueError(
            f"""
            func can't squeeze input dimensions. If x has shape (d_1, ..., d_i), func(x) must have shape (..., d_1, ..., d_i).
            Ex : if x.shape == (m, n), func(x).shape == (..., m, n).
            """
        )

    exp_a = np.where(np.exp(-arr_a) == 0, 1.0, np.exp(-arr_a))  # () or (n,) or (m, n)
    return np.sum(
        w * fvalues * exp_a, axis=-shifted_x.ndim
    )  # (d_1, ..., d_i) or (d_1, ..., d_i, n) or (d_1, ..., d_i, m, n)


def unweighted_laguerre_quadrature(
    func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
    a: AnyFloat,
    deg: int = 10,
):
    r"""Numerical integration of :math:`f(x)` over the interval :math:`[a, \infty]`

    `func` must accept (deg,), (deg, n) or (deg, m, n) array shapes
    It must handle at least 3 dimensions.
    a can be zero with ndim <= 2.
    """

    arr_a = broadcast_bounds(a)  # () or (n,) or (m, n)
    quad = np.polynomial.laguerre.laggauss(deg)  # (deg,)
    x, w = quad
    x = np.expand_dims(
        x, axis=tuple(range(1, arr_a.ndim + 1))
    )  # (deg,), (deg, 1) or (deg, 1, 1)
    w = np.expand_dims(
        w, axis=tuple(range(1, arr_a.ndim + 1))
    )  # (deg,), (deg, 1) or (deg, 1, 1)

    shifted_x = x + arr_a  # (deg,) or (deg, n) or (deg, m, n)
    fvalues = func(
        shifted_x
    )  # (d_1, ..., d_i, deg) or (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
    try:
        _ = np.broadcast_shapes(fvalues.shape[-len(shifted_x.shape) :], shifted_x.shape)
    except ValueError:
        raise ValueError(
            f"""
            func can't squeeze input dimensions. If x has shape (d_1, ..., d_i), func(x) must have shape (..., d_1, ..., d_i).
            Ex : if x.shape == (m, n), func(x).shape == (..., m, n).
            """
        )
    return np.sum(
        w * fvalues * np.exp(x), axis=-shifted_x.ndim
    )  # (d_1, ..., d_i) or (d_1, ..., d_i, n) or (d_1, ..., d_i, m, n)



# pyright: basic

from abc import ABC, abstractmethod
from collections.abc import Iterator

import numpy as np
from numpy.lib import recfunctions as rfn
from numpy.typing import NDArray
from typing_extensions import override

from relife.economic import ExponentialDiscounting, Reward
from relife.lifetime_model import LeftTruncatedModel
from relife.lifetime_model._base import ParametricLifetimeModel
from relife.lifetime_model._distribution import (
    EquilibriumDistribution,
)
from relife.base import is_frozen

__all__ = [
    "StochasticDataIterator",
    "RenewalProcessIterator",
    "RenewalRewardProcessIterator",
    "NonHomogeneousPoissonProcessIterator",
]


def _expand_lifetime_model(
    lifetime_model: ParametricLifetimeModel, nb_samples: int
) -> ParametricLifetimeModel:
    """
    Expand a lifetime model by duplicating its arguments. A regression with n_assets assets will result in a new regression with n_assets * n_samples assets.
    """
    if isinstance(lifetime_model, EquilibriumDistribution):
        return EquilibriumDistribution(
            _expand_lifetime_model(lifetime_model.baseline, nb_samples)
        )

    expanded_lifetime_model = lifetime_model

    if is_frozen(lifetime_model):
        broadcasted_args = list(
            np.repeat(arg, nb_samples, axis=0) for arg in lifetime_model.args
        )
        expanded_lifetime_model = lifetime_model.unfreeze().freeze(
            *broadcasted_args
        )  # TODO: use a copy method of parametric models

    return expanded_lifetime_model


class StochasticDataIterator(Iterator[NDArray[np.void]], ABC):
    """Abstract class for all stochastic processes iterator.
    Used to build the structarrays, get the shapes, iterate through steps and identify the observation window.
    Abstract method is sample_next_step, that is unique for each stochastic process.
    """

    def __init__(
        self,
        process,
        nb_samples: int,
        time_window: tuple[float, float],
        nb_assets: int = 1,
        seed=None,
    ) -> None:
        self.process = process
        self.nb_samples = nb_samples
        self.nb_assets = nb_assets

        self.sample_size = self.nb_assets * self.nb_samples

        self._expanded_lifetime_model = _expand_lifetime_model(
            self.process.lifetime_model, self.nb_samples
        )

        self.t0, self.tf = time_window  # t0, tf are checked in iterable

        self.seed = np.random.default_rng(seed)

        # Assets ages restart at 0 when replacements are done.
        # Used to identify the current ages of the assets in the process
        self.ages = np.zeros(self.sample_size, dtype=np.float64)
        # Full timeline never restarts, sums at each iteration
        # Used to identify current time for the observer
        self.timeline = np.zeros(self.sample_size, dtype=np.float64)

        self.replacement_cycle = 0

        self._crossed_t0_counter = np.zeros(self.sample_size, dtype=np.uint32)
        self._crossed_tf_counter = np.zeros(self.sample_size, dtype=np.uint32)

        self._time_window_mask = np.zeros(
            self.sample_size,
            dtype=np.dtype(
                [
                    ("observed_step", np.bool_),
                    ("just_crossed_t0", np.bool_),
                    ("just_crossed_tf", np.bool_),
                    ("crossed_tf", np.bool_),
                ]
            ),
        )

    @property
    @abstractmethod
    def _expanded_dynamic_lifetime_model(self) -> ParametricLifetimeModel:
        """
        Use the lifetime model modified at each iteration according to each stochastic process specific properties
        """

    @abstractmethod
    def update_ages(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.bool_],
        entry: NDArray[np.float64],
    ) -> None:
        """
        Update ages at each iteration according to each stochastic process specific properties
        """

    def sample_time_event_entry(
        self,
    ) -> tuple[NDArray[np.float64], NDArray[np.bool_], NDArray[np.float64]]:
        """
        Routine that samples time, event, entry at each step
        IMPORTANT: time, event, entry must be 1D.
        """

        time, event, entry = self._expanded_dynamic_lifetime_model.rvs(
            (self.sample_size, 1),
            return_event=True,
            return_entry=True,
            seed=self.seed,
        )

        time = time.flatten()
        event = event.flatten()
        entry = entry.flatten()

        return time, event, entry

    def __next__(self) -> NDArray[np.void]:
        """function to iterate"""
        if not np.all(self._time_window_mask["crossed_tf"]):
            time, event, entry = self.sample_time_event_entry()
            struct_arr = self._collect_time_window_observations(time, event, entry)
            self.update_ages(time, event, entry)
            while (
                struct_arr.size == 0
            ):  # skip cycles while arrays are empty (if t0 != 0.)
                time, event, entry = self.sample_time_event_entry()
                struct_arr = self._collect_time_window_observations(time, event, entry)
                self.update_ages(time, event, entry)
                if np.all(self._time_window_mask["crossed_tf"]) and struct_arr.size > 0:
                    return struct_arr
            return struct_arr
        raise StopIteration

    def _collect_time_window_observations(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.bool_],
        entry: NDArray[np.float64],
    ) -> NDArray[np.void]:
        """Collect observed time, event, entry inside during the time window"""
        if time.ndim > 1:
            raise ValueError(
                f"sample_time_event_entry must return 1d entry. Got {time.ndim} dim for time"
            )
        if event.ndim > 1:
            raise ValueError(
                f"sample_time_event_entry must return 1d entry. Got {event.ndim} dim for event"
            )
        if entry.ndim > 1:
            raise ValueError(
                f"sample_time_event_entry must return 1d entry. Got {entry.ndim} dim for entry"
            )

        residual_time = time - entry

        # Timeline increases by residual time
        self.timeline += residual_time
        self._update_time_window_mask()

        # Replace times that exceeds tf with right censorings at tf
        time = np.where(
            self._time_window_mask["just_crossed_tf"],
            time - (self.timeline - self.tf),
            time,
        )
        self.timeline[self._time_window_mask["just_crossed_tf"]] = self.tf
        event[self._time_window_mask["just_crossed_tf"]] = False

        # Replace entries to take account of the left truncation at t0
        entry = np.where(
            self._time_window_mask["just_crossed_t0"],
            time - (self.timeline - self.t0),
            entry,
        )

        self.replacement_cycle += 1

        struct_arr = rfn.append_fields(  #  works on structured_array too
            self._init_structarray(),
            ("time", "event", "entry"),
            (
                time[self._time_window_mask["observed_step"]],
                event[self._time_window_mask["observed_step"]],
                entry[self._time_window_mask["observed_step"]],
            ),
            (np.float64, np.bool_, np.float64),
            usemask=False,
            asrecarray=False,
        )
        return struct_arr

    def _update_time_window_mask(self) -> None:
        """Update mask to keep track of the observed and non observed events"""
        self._crossed_t0_counter[self.timeline > self.t0] += 1
        self._crossed_tf_counter[self.timeline > self.tf] += 1
        self._time_window_mask["just_crossed_t0"] = self._crossed_t0_counter == 1
        self._time_window_mask["just_crossed_tf"] = self._crossed_tf_counter == 1
        self._time_window_mask["crossed_tf"] = self._crossed_tf_counter >= 1
        self._time_window_mask["observed_step"] = np.logical_and(
            self._crossed_t0_counter >= 1, self._crossed_tf_counter <= 1
        )

    def _init_structarray(self) -> NDArray[np.void]:
        """Construct the struct array to return"""
        observed_index = np.where(self._time_window_mask["observed_step"])[0]

        asset_id = observed_index // self.nb_samples
        sample_id = observed_index % self.nb_samples

        struct_array = np.zeros(
            sample_id.size,
            dtype=np.dtype(
                [
                    ("asset_id", np.uint32),  #  unsigned 32bit integer
                    ("sample_id", np.uint32),  #  unsigned 32bit integer
                    ("timeline", np.float64),  #  64bit float
                ]
            ),
        )

        struct_array["asset_id"] = asset_id.astype(np.uint32)
        struct_array["sample_id"] = sample_id.astype(np.uint32)
        struct_array["timeline"] = self.timeline[
            self._time_window_mask["observed_step"]
        ]
        return struct_array


class RenewalProcessIterator(StochasticDataIterator):
    def __init__(
        self,
        process,
        nb_samples: int,
        time_window: tuple[float, float],
        nb_assets: int = 1,
        seed=None,
    ) -> None:
        super().__init__(
            process=process,
            nb_samples=nb_samples,
            time_window=time_window,
            nb_assets=nb_assets,
            seed=seed,
        )

        self._expanded_first_lifetime_model = None
        if self.process.first_lifetime_model is not None:
            self._expanded_first_lifetime_model = _expand_lifetime_model(
                self.process.first_lifetime_model, self.nb_samples
            )

    @property
    def _expanded_dynamic_lifetime_model(self) -> ParametricLifetimeModel:
        if (
            self.replacement_cycle == 0
            and self._expanded_first_lifetime_model is not None
        ):
            return self._expanded_first_lifetime_model
        return self._expanded_lifetime_model

    def update_ages(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.bool_],
        entry: NDArray[np.float64],
    ):
        """
        In a Renewal process, ages are reset to 0 after each iteration. The ages array remains constant.
        """
        pass


class RenewalRewardProcessIterator(RenewalProcessIterator):
    reward: Reward
    discounting: ExponentialDiscounting

    def __init__(
        self,
        process,
        nb_samples: int,
        time_window: tuple[float, float],
        nb_assets: int = 1,
        seed: int | None = None,
    ) -> None:
        super().__init__(
            process, nb_samples, time_window, nb_assets=nb_assets, seed=seed
        )
        self.reward = self.process.reward
        self.discounting = self.process.discounting

    @override
    def __next__(self) -> NDArray[np.void]:
        struct_array = super().__next__()
        # may be type hint error in rfn.append_fields overload
        return rfn.append_fields(
            struct_array,
            "reward",
            self.reward.sample(struct_array["time"])
            * self.discounting.factor(struct_array["timeline"]),
            np.float64,
            usemask=False,
            asrecarray=False,
        )  # type: ignore


class StochasticAgesProcessIterator(StochasticDataIterator):
    """
    Abstract class for all stochastic process iterators that involve an age or virtual age for the iterative sampling
    """

    @property
    def _expanded_dynamic_lifetime_model(self) -> ParametricLifetimeModel:
        # Apply a Left truncation based on current ages on the model
        # self.ages is always 1d in LeftTruncatedModel
        ages = self.ages.copy()
        return LeftTruncatedModel(self._expanded_lifetime_model).freeze(ages)


class NonHomogeneousPoissonProcessIterator(StochasticAgesProcessIterator):
    def update_ages(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.bool_],
        entry: NDArray[np.float64],
    ):
        """
        In a NHPP, ages are reset to 0 only when a replacement is made
        """
        # Update asset ages
        self.ages += time - entry
        # If no events (replacement), restarts age for next step
        self.ages[~event] = 0


class Kijima1ProcessIterator(StochasticAgesProcessIterator):
    def update_ages(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.bool_],
        entry: NDArray[np.float64],
    ):
        """
        In a Kijima Process, the concept of age is virtual, and depends on the q parameter of the process
        """
        # Update asset ages
        self.ages = self.ages + self.process.q * (time - entry)
        # If no events (replacement), restarts age for next step
        self.ages[~event] = 0


class Kijima2ProcessIterator(StochasticAgesProcessIterator):
    def update_ages(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.bool_],
        entry: NDArray[np.float64],
    ):
        """
        In a Kijima Process, the concept of age is virtual, and depends on the q parameter of the process
        """
        # Update asset ages
        self.ages = self.process.q * (self.ages + time - entry)
        # If no events (replacement), restarts age for next step
        self.ages[~event] = 0



# pyright: basic

import numpy as np
import pytest
from pytest import approx


def rvs_expected_shape(size):
    if size != 1:
        return (size,)
    return ()


def test_rvs(distribution, rvs_size):
    assert distribution.rvs(
        rvs_size,
    ).shape == rvs_expected_shape(
        size=rvs_size,
    )
    assert all(
        arr.shape
        == rvs_expected_shape(
            size=rvs_size,
        )
        for arr in distribution.rvs(rvs_size, return_event=True)
    )
    assert all(
        arr.shape
        == rvs_expected_shape(
            size=rvs_size,
        )
        for arr in distribution.rvs(rvs_size, return_entry=True)
    )
    assert all(
        arr.shape
        == rvs_expected_shape(
            size=rvs_size,
        )
        for arr in distribution.rvs(rvs_size, return_event=True, return_entry=True)
    )


def test_sf(distribution, time):
    assert distribution.sf(time).shape == time.shape
    assert distribution.sf(np.full(time.shape, distribution.median())) == approx(
        np.full(time.shape, 0.5), rel=1e-3
    )


def test_hf(distribution, time):
    assert distribution.hf(time).shape == time.shape


def test_chf(distribution, time):
    assert distribution.chf(time).shape == time.shape


def test_cdf(distribution, time):
    assert distribution.cdf(time).shape == time.shape


def test_pdf(distribution, time):
    assert distribution.pdf(time).shape == time.shape


def test_ppf(distribution, probability):
    assert distribution.ppf(probability).shape == probability.shape


def test_ichf(distribution, probability):
    assert distribution.ichf(probability).shape == probability.shape


def test_isf(distribution, probability):
    assert distribution.isf(probability).shape == probability.shape
    assert distribution.isf(np.full(probability.shape, 0.5)) == approx(
        np.full(probability.shape, distribution.median())
    )


def test_moment(distribution):
    assert distribution.moment(1).shape == ()
    assert distribution.moment(2).shape == ()


def test_mean(distribution):
    assert distribution.mean().shape == ()


def test_var(distribution):
    assert distribution.var().shape == ()


def test_median(distribution):
    assert distribution.median().shape == ()


def test_dhf(distribution, time):
    assert distribution.dhf(time).shape == time.shape


def test_jac_sf(distribution, time):
    assert distribution.jac_sf(time).shape == (distribution.nb_params,) + time.shape


def test_jac_hf(distribution, time):
    assert distribution.jac_hf(time).shape == (distribution.nb_params,) + time.shape


def test_jac_chf(distribution, time):
    assert distribution.jac_chf(time).shape == (distribution.nb_params,) + time.shape


def test_jac_cdf(distribution, time):
    assert distribution.jac_cdf(time).shape == (distribution.nb_params,) + time.shape


def test_jac_pdf(distribution, time):
    assert distribution.jac_pdf(time).shape == (distribution.nb_params,) + time.shape


def test_ls_integrate(distribution, integration_bound_a, integration_bound_b):
    # integral_a^b dF(x)
    expected_shape = np.broadcast_shapes(
        integration_bound_a.shape, integration_bound_b.shape
    )
    integration = distribution.ls_integrate(
        np.ones_like, integration_bound_a, integration_bound_b, deg=100
    )
    assert integration.shape == expected_shape
    assert integration == approx(
        distribution.cdf(integration_bound_b) - distribution.cdf(integration_bound_a)
    )
    # integral_0^inf x*dF(x)
    integration = distribution.ls_integrate(
        lambda x: x,
        np.zeros_like(integration_bound_a),
        np.full_like(integration_bound_b, np.inf),
        deg=100,
    )
    assert integration == approx(np.full(expected_shape, distribution.mean()), rel=1e-3)


def test_fit(distribution, power_transformer_data):
    expected_params = distribution.params.copy()
    distribution = distribution.fit(
        power_transformer_data["time"],
        event=power_transformer_data["event"],
        entry=power_transformer_data["entry"],
    )
    assert distribution.params == pytest.approx(expected_params, rel=1e-3)


def test_negative_log(distribution_likelihood):
    params = distribution_likelihood.model.params.copy()
    assert isinstance(distribution_likelihood.negative_log(params), float)


def test_jac_negative_log(distribution_likelihood):
    params = distribution_likelihood.model.params.copy()
    assert distribution_likelihood.jac_negative_log(params).shape == (params.size,)


class TestEquilibriumDistribution:
    # def test_args_names(self, equilibrium_distribution):
    #     assert equilibrium_distribution.args_names == ()
    #
    # def test_rvs(self, equilibrium_distribution):
    #     m, n = 3, 10
    #     assert equilibrium_distribution.rvs(seed=21).shape == ()
    #     assert equilibrium_distribution.rvs(size=(n,), seed=21).shape == (n,)
    #     assert equilibrium_distribution.rvs(size=(m, 1), seed=21).shape == (m, 1)
    #     assert equilibrium_distribution.rvs(size=(m, n), seed=21).shape == (m, n)

    @pytest.mark.xfail
    def test_moment(self, equilibrium_distribution):
        assert equilibrium_distribution.moment(1).shape == ()
        assert equilibrium_distribution.moment(2).shape == ()

    @pytest.mark.xfail
    def test_mean(self, equilibrium_distribution):
        assert equilibrium_distribution.mean().shape == ()

    @pytest.mark.xfail
    def test_var(self, equilibrium_distribution):
        assert equilibrium_distribution.var().shape == ()

    @pytest.mark.xfail
    def test_median(self, equilibrium_distribution):
        assert equilibrium_distribution.median().shape == ()



# pyright: basic
from __future__ import annotations

import warnings
from abc import ABC
from typing import Any, Literal, Optional, Self, overload

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import newton

from relife.economic import AgeReplacementReward, ExponentialDiscounting
from relife.lifetime_model import (
    AgeReplacementModel,
    LeftTruncatedModel,
)
from relife.stochastic_process import RenewalRewardProcess
from relife.stochastic_process.non_homogeneous_poisson_process import (
    FrozenNonHomogeneousPoissonProcess,
)
from relife.typing import (
    AnyFloat,
    AnyParametricLifetimeModel,
    NumpyFloat,
)
from relife.utils import (
    flatten_if_possible,
    reshape_1d_arg,
)
from relife.utils.quadrature import legendre_quadrature

from ._base import OneCycleExpectedCosts, ReplacementPolicy

__all__ = [
    "OneCycleAgeReplacementPolicy",
    "AgeReplacementModel",
    "age_replacement_policy",
]


@overload
def age_replacement_policy(
    baseline_model: AnyParametricLifetimeModel[()],
    cost_structure: dict[str, AnyFloat],
    one_cycle: Literal[True],
    **kwargs: Any,
) -> OneCycleAgeReplacementPolicy: ...
@overload
def age_replacement_policy(
    baseline_model: AnyParametricLifetimeModel[()],
    cost_structure: dict[str, AnyFloat],
    one_cycle: Literal[False],
    **kwargs: Any,
) -> AgeReplacementPolicy: ...
@overload
def age_replacement_policy(
    baseline_model: FrozenNonHomogeneousPoissonProcess[*tuple[Any, ...]],
    cost_structure: dict[str, AnyFloat],
    one_cycle: bool = False,
    **kwargs: Any,
) -> NonHomogeneousPoissonAgeReplacementPolicy: ...
@overload
def age_replacement_policy(
    baseline_model: AnyParametricLifetimeModel[()]
    | FrozenNonHomogeneousPoissonProcess[*tuple[Any, ...]],
    cost_structure: dict[str, AnyFloat],
    one_cycle: bool = False,
    **kwargs: Any,
) -> (
    OneCycleAgeReplacementPolicy
    | AgeReplacementPolicy
    | NonHomogeneousPoissonAgeReplacementPolicy
): ...
def age_replacement_policy(
    baseline_model: AnyParametricLifetimeModel[()]
    | FrozenNonHomogeneousPoissonProcess[*tuple[Any, ...]],
    cost_structure: dict[str, AnyFloat],
    one_cycle: bool = False,
    **kwargs: Any,
) -> (
    OneCycleAgeReplacementPolicy
    | AgeReplacementPolicy
    | NonHomogeneousPoissonAgeReplacementPolicy
):
    """
    Creates a preventive age replacement policy.

    Parameters
    ----------
    baseline_model : parametric model
        Parametric model required by the policy.
    cost_structure : dict of costs
        Dictionnary containing the cost values (float or 1d-array) and their corresponding names (either cf, cp or cr).
    one_cycle : bool, default False
        If True, returns the one cycle variation of the policy.
    **kwargs
        Extra arguments required by the policy (a0, ar, discounting_rate, etc.)

    Returns
    -------
    Policy
        Policy corresponding to the ``baseline_model`` and the ``cost_structure``.

    Raises
    ------
    ValueError
        If ``baseline_model`` or ``cost_structure`` does not have a corresponding policy.
    """
    if isinstance(baseline_model, FrozenNonHomogeneousPoissonProcess):
        try:
            cr = cost_structure["cr"]
            cp = cost_structure["cp"]
        except KeyError:
            raise ValueError("costs must contain 'cr' and 'cp'")
        return NonHomogeneousPoissonAgeReplacementPolicy(
            baseline_model, cr, cp, **kwargs
        )
    try:
        cf = cost_structure["cf"]
        cp = cost_structure["cp"]
    except KeyError:
        raise ValueError("costs must contain 'cf' and 'cp'")
    if one_cycle:
        return OneCycleAgeReplacementPolicy(baseline_model, cf, cp, **kwargs)
    return AgeReplacementPolicy(baseline_model, cf, cp, **kwargs)


class BaseAgeReplacementPolicy(ReplacementPolicy[AnyParametricLifetimeModel[()]], ABC):
    _cost_structure: dict[str, NumpyFloat]
    _ar: Optional[NumpyFloat]
    _a0: Optional[NumpyFloat]
    discounting_rate: float

    def __init__(
        self,
        lifetime_model: AnyParametricLifetimeModel[()],
        cf: AnyFloat,
        cp: AnyFloat,
        discounting_rate: float = 0.0,
        a0: Optional[AnyFloat] = None,
        ar: Optional[AnyFloat] = None,
    ):
        super().__init__(
            lifetime_model,
            {"cf": reshape_1d_arg(cf), "cp": reshape_1d_arg(cp)},
            discounting_rate=discounting_rate,
        )
        self._a0 = reshape_1d_arg(a0) if a0 is not None else a0
        self._ar = reshape_1d_arg(ar) if ar is not None else ar

    @property
    def a0(self) -> Optional[NumpyFloat]:
        """Current ages of the assets.

        Returns
        -------
        np.ndarray
        """
        # _a0 is (m, 1) but exposed cf is (m,)
        if self._a0 is None:
            return self._a0
        return flatten_if_possible(self._a0)

    @property
    def cf(self) -> NumpyFloat:
        """Cost of failure.

        Returns
        -------
        np.ndarray
        """
        # _cf is (m, 1) but exposed cf is (m,)
        return flatten_if_possible(self._cost_structure["cf"])

    @cf.setter
    def cf(self, value: AnyFloat) -> None:
        self._cost_structure["cf"] = reshape_1d_arg(value)

    @property
    def cp(self) -> NumpyFloat:
        """Costs of preventive replacements.

        Returns
        -------
        np.ndarray
        """
        # _cp is (m, 1) but exposed cp is (m,)
        return flatten_if_possible(self._cost_structure["cp"])

    @cp.setter
    def cp(self, value: AnyFloat) -> None:
        self._cost_structure["cp"] = reshape_1d_arg(value)

    @property
    def ar(self) -> Optional[NumpyFloat]:
        """Preventive ages of replacement.

        Returns
        -------
        np.ndarray
        """
        # _ar is (m, 1) but exposed ar is (m,)
        if self._ar is None:
            return self._ar
        return flatten_if_possible(self._ar)

    @ar.setter
    def ar(self, value: Optional[AnyFloat]) -> None:
        if value is not None:
            self._ar = reshape_1d_arg(value)
        else:
            self._ar = None

    @property
    def tr1(self) -> Optional[NumpyFloat]:
        """Times before the first replacement.

        Returns
        -------
        np.ndarray
        """
        if self._a0 is not None and self._ar is not None:
            tr = np.maximum(self._ar - self._a0, 0)
            return flatten_if_possible(tr)
        return self.ar


class OneCycleAgeReplacementPolicy(BaseAgeReplacementPolicy):
    r"""One-cyle age replacement policy.

    Asset is replaced at a fixed age :math:`a_r` with cost :math:`c_p` or it is replaced
    upon failure with cost :math:`c_f`.

    .. note::

        ``OneCycleAgeReplacementPolicy`` differs from ``AgeReplacementPolicy``
        because only one cycle of replacement is considered.

    The object's methods require the ``ar`` attribute to be set either at the instanciation
    or by calling the ``optimize`` method. Otherwise, an error will be raised.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    cf : float or 1darray
        Costs of failures
    cp : float or 1darray
        Costs of preventive replacements
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    a0 : float or 1darray, optional
        Current ages of the assets. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.
    ar : float or 1darray, optional
        Ages of preventive replacements, by default None. If not given, one must call ``optimize`` to set ``ar`` values
        and access to the rest of the object interface.

    Attributes
    ----------
    cf
    cp
    ar

    References
    ----------
    .. [1] Coolen-Schrijner, P., & Coolen, F. P. A. (2006). On optimality
        criteria for age replacement. Proceedings of the Institution of
        Mechanical Engineers, Part O: Journal of Risk and Reliability,
        220(1), 21-29
    """

    period_before_discounting: float

    def __init__(
        self,
        lifetime_model: AnyParametricLifetimeModel[()],
        cf: AnyFloat,
        cp: AnyFloat,
        discounting_rate: float = 0.0,
        period_before_discounting: float = 1.0,
        a0: Optional[AnyFloat] = None,
        ar: Optional[AnyFloat] = None,
    ):
        super().__init__(
            lifetime_model, cf, cp, discounting_rate=discounting_rate, a0=a0, ar=ar
        )
        self.period_before_discounting = period_before_discounting

    @property
    def _expected_costs(self) -> OneCycleExpectedCosts:
        if self.ar is None or self.tr1 is None:
            raise ValueError("ar must be set or optimized")
        else:
            if self.a0 is None:
                return OneCycleExpectedCosts(
                    AgeReplacementModel(self.baseline_model).freeze(self.tr1),
                    AgeReplacementReward(self.cf, self.cp, self.tr1),
                    discounting_rate=self.discounting_rate,
                    period_before_discounting=self.period_before_discounting,
                )
            return OneCycleExpectedCosts(
                AgeReplacementModel(LeftTruncatedModel(self.baseline_model)).freeze(
                    self.tr1, self.a0
                ),
                AgeReplacementReward(self.cf, self.cp, self.tr1),
                discounting_rate=self.discounting_rate,
                period_before_discounting=self.period_before_discounting,
            )

    def expected_net_present_value(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        if self.ar is None:
            raise ValueError
        return self._expected_costs.expected_net_present_value(
            tf, nb_steps, total_sum=total_sum
        )  # (nb_steps,), (nb_steps,) or (nb_steps,), (m, nb_steps)

    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        if self.ar is None:
            raise ValueError
        return (
            self._expected_costs.asymptotic_expected_net_present_value()
        )  # () or (m, nb_steps)

    def expected_equivalent_annual_cost(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        if self.ar is None or self.tr1 is None:
            raise ValueError
        else:
            # because b = np.minimum(ar, b) in ls_integrate, b can be lower than a depending on period before discounting
            if np.any(self.period_before_discounting >= self.ar):
                raise ValueError(
                    "The period before discounting must be lower than ar values"
                )

            if np.any(self.tr1 == 0):
                warnings.warn(
                    "Some assets has already been replaced for the first cycle (where tr is 0). For these assets, consider adjusting ar values to be greater than a0"
                )

            ar = self.ar.copy()
            #  change ar temporarly to enable computation of eeac (if not, AgeReplacementModel.ls_integrate bounds will be problematic)
            self.ar = np.where(self.tr1 == 0, np.inf, self.ar)
            timeline, eeac = self._expected_costs.expected_equivalent_annual_cost(
                tf, nb_steps
            )
            if eeac.ndim == 2:  # more than one asset
                eeac[np.where(self.ar == np.inf)[0], :] = np.nan
            if eeac.ndim == 1 and self.ar == np.inf:
                eeac.fill(np.nan)
            self.ar = ar
            return (
                timeline,
                eeac,
            )  # (nb_steps,), (nb_steps,) or (nb_steps,), (m, nb_steps)

    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        if self.ar is None:
            raise ValueError
        # because b = np.minimum(ar, b) in ls_integrate, b can be lower than a depending on period before discounting
        if np.any(self.period_before_discounting >= self.ar):
            raise ValueError(
                "The period before discounting must be lower than ar values"
            )

        if np.any(self.tr1 == 0):
            warnings.warn(
                "Some assets has already been replaced for the first cycle (where tr is 0). For these assets, consider adjusting ar values to be greater than a0"
            )

        ar = self.ar.copy()
        #  change ar temporarly to enable computation of eeac (if not, AgeReplacementModel.ls_integrate bounds are not in right order)
        self.ar = np.where(self.tr1 == 0, np.inf, self.ar)
        asymptotic_eeac = (
            self._expected_costs.asymptotic_expected_equivalent_annual_cost()
        )

        if asymptotic_eeac.ndim == 1:  # more than one asset
            asymptotic_eeac[np.where(self.ar == np.inf)[0]] = np.nan
        if asymptotic_eeac.ndim == 0 and self.ar == np.inf:
            asymptotic_eeac = np.array(np.nan)
        self.ar = ar
        return asymptotic_eeac  # () or (m, nb_steps)

    def optimize(self) -> Self:
        """
        Optimize the policy according the costs, the discounting rate and the underlying lifetime model.

        Returns
        -------
        Self
             Optimized policy.
        """

        discounting = ExponentialDiscounting(self.discounting_rate)

        x0 = np.minimum(
            self._cost_structure["cp"]
            / (self._cost_structure["cf"] - self._cost_structure["cp"]),
            1,
        )  # () or (m, 1)

        # if costs are floats, x0 is float. BUT, baseline_model can have many assets
        # x0 must have the same shape than eq(x0) (see scipy.newton doc)
        _sf_x0 = self.baseline_model.sf(x0)  # gives us the shape, thus nb_assets
        if _sf_x0.ndim == 2 and x0.ndim == 0:
            x0 = np.tile(x0, (_sf_x0.shape[0], 1))

        def eq(a: NDArray[np.float64]) -> NDArray[np.float64]:
            return (
                discounting.factor(a)
                / discounting.annuity_factor(a)
                * (
                    (self._cost_structure["cf"] - self._cost_structure["cp"])
                    * self.baseline_model.hf(a)
                    - self._cost_structure["cp"] / discounting.annuity_factor(a)
                )
            )

        # no idea on how to type eq
        self.ar = newton(eq, x0)  # () or (m, 1) # pyright:ignore
        return self


class AgeReplacementPolicy(BaseAgeReplacementPolicy):
    r"""Age replacement policy.

    Asset is replaced at a fixed age :math:`a_r` with cost :math:`c_p` or it is replaced
    upon failure with cost :math:`c_f`.

    The object's methods require the ``ar`` attribute to be set either at the instanciation
    or by calling the ``optimize`` method. Otherwise, an error will be raised.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    cf : float or 1darray
        Costs of failures
    cp : float or 1darray
        Costs of preventive replacements
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    a0 : float or 1darray, optional
        Current ages of the assets. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.
    ar : float or 1darray, optional
        Ages of preventive replacements, by default None. If not given, one must call ``optimize`` to set ``ar`` values
        and access to the rest of the object interface.

    Attributes
    ----------
    cf
    cp
    ar

    References
    ----------
    .. [1] Mazzuchi, T. A., Van Noortwijk, J. M., & Kallen, M. J. (2007).
        Maintenance optimization. Encyclopedia of Statistics in Quality and
        Reliability, 1000-1008.
    """

    @property
    def _stochastic_process(self) -> RenewalRewardProcess:
        if self.ar is None or self.tr1 is None:
            raise ValueError("ar must be set or optimized")
        if self.a0 is None:
            return RenewalRewardProcess(
                AgeReplacementModel(self.baseline_model).freeze(self.ar),
                AgeReplacementReward(self.cf, self.cp, self.ar),
                discounting_rate=self.discounting_rate,
            )
        return RenewalRewardProcess(
            AgeReplacementModel(self.baseline_model).freeze(self.ar),
            AgeReplacementReward(self.cf, self.cp, self.ar),
            discounting_rate=self.discounting_rate,
            first_lifetime_model=AgeReplacementModel(
                LeftTruncatedModel(self.baseline_model)
            ).freeze(self.tr1, self.a0),
            first_reward=AgeReplacementReward(self.cf, self.cp, self.tr1),
        )

    def expected_net_present_value(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        if self.ar is None:
            raise ValueError("ar must be set or optimized")
        timeline, npv = self._stochastic_process.expected_total_reward(tf, nb_steps)
        if total_sum and npv.ndim == 2:
            npv = np.sum(npv, axis=0)
        return timeline, npv

    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        asymptotic_npv = (
            self._stochastic_process.asymptotic_expected_total_reward()
        )  # () or (m,)
        if total_sum:
            asymptotic_npv = np.sum(asymptotic_npv)
        return asymptotic_npv

    def expected_equivalent_annual_cost(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        if self.ar is None or self.tr1 is None:
            raise ValueError

        timeline, eeac = self._stochastic_process.expected_equivalent_annual_worth(
            tf, nb_steps
        )
        if np.any(self.tr1 == 0):
            warnings.warn(
                "Some assets has already been replaced for the first cycle (where tr1 is 0). For these assets, consider adjusting ar values to be greater than a0"
            )
        if eeac.ndim == 2:
            eeac[np.where(np.atleast_1d(self.tr1) == 0)] = np.nan
        if eeac.ndim == 1 and self.ar == np.inf:
            eeac.fill(np.nan)
        if total_sum and eeac.ndim == 2:
            eeac = np.sum(eeac, axis=0)
        return timeline, eeac  # (nb_steps,), (nb_steps,) or (nb_steps,), (m, nb_steps)

    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        if self.ar is None or self.tr1 is None:
            raise ValueError

        asymptotic_eeac = (
            self._stochastic_process.asymptotic_expected_equivalent_annual_worth()
        )
        if np.any(self.tr1 == 0):
            warnings.warn(
                "Some assets has already been replaced for the first cycle (where tr1 is 0). For these assets, consider adjusting ar values to be greater than a0"
            )
        if asymptotic_eeac.ndim == 1:
            asymptotic_eeac[np.where(self.tr1 == 0)] = np.nan
        if asymptotic_eeac.ndim == 0 and self.tr1 == 0:
            asymptotic_eeac = np.array(np.nan)
        if total_sum:
            asymptotic_eeac = np.sum(asymptotic_eeac)
        return asymptotic_eeac  # () or (m,)

    def annual_number_of_replacements(self, nb_years, upon_failure=False, total=True):
        """
        The expected number of annual replacements.

        Parameters
        ----------
        nb_years : int
            The number of years on which the annual number of replacements are projected
        upon_failure : bool, default is False
            If True, it also returns the annual number of replacements due to unexpected failures
        total : bool, default is True
            If True, the given numbers of replacements are the sum of all replacements without distinction between assets
        """
        copied_policy = self.__class__(
            self.baseline_model,
            1.0,
            1.0,
            ar=self.ar,
            a0=self.a0,
            discounting_rate=0.0,
        )
        timeline, total_cost = copied_policy.expected_net_present_value(
            nb_years, nb_years + 1
        )  # equiv to np.arange
        if total:
            mt = np.sum(np.atleast_2d(total_cost), axis=0)
        else:
            mt = total_cost
        nb_replacements = np.diff(mt)
        if upon_failure:
            copied_policy = self.__class__(
                self.baseline_model,
                1.0,
                0.0,
                ar=self.ar,
                a0=self.a0,
                discounting_rate=0.0,
            )
            _, total_cost = copied_policy.expected_net_present_value(
                nb_years, nb_years + 1
            )  # equiv to np.arange
            if total:
                mf = np.sum(np.atleast_2d(total_cost), axis=0)
            else:
                mf = total_cost
            nb_failures = np.diff(mf)
            return timeline[1:], nb_replacements, nb_failures
        return timeline[1:], nb_replacements

    def optimize(self) -> Self:
        """
        Optimize the policy according the costs, the discounting rate and the underlying lifetime model.

        Returns
        -------
        Self
             Optimized policy.
        """
        discounting = ExponentialDiscounting(self.discounting_rate)
        x0 = np.minimum(
            self._cost_structure["cp"]
            / (self._cost_structure["cf"] - self._cost_structure["cp"]),
            1,
        )

        # if costs are floats, x0 is float. BUT, baseline_model can have many assets
        # x0 must have the same shape than eq(x0) (see scipy.newton doc)
        _sf_x0 = self.baseline_model.sf(x0)  # gives us the shape, thus nb_assets
        if _sf_x0.ndim == 2 and x0.ndim == 0:
            x0 = np.tile(x0, (_sf_x0.shape[0], 1))

        def eq(a: NDArray[np.float64]) -> NDArray[np.float64]:  # () or (m, 1)
            f = legendre_quadrature(
                lambda x: discounting.factor(x) * self.baseline_model.sf(x),
                0,
                a,
            )
            g = legendre_quadrature(
                lambda x: discounting.factor(x) * self.baseline_model.pdf(x),
                0,
                a,
            )
            return (
                discounting.factor(a)
                * (
                    (self._cost_structure["cf"] - self._cost_structure["cp"])
                    * (self.baseline_model.hf(a) * f - g)
                    - self._cost_structure["cp"]
                )
                / f**2
            )

        # no idea on how to type eq properly
        self.ar = newton(eq, x0)  # () or (m, 1)  # pyright: ignore
        return self

    def generate_failure_data(
        self, nb_samples: int, time_window: tuple[float, float], seed=None
    ):
        """Generate failure data

        This function will generate failure data that can be used to fit a lifetime model.

        Parameters
        ----------
        nb_samples : int
            The number of samples.
        time_window : tuple of two floats
            Time window in which data are sampled.
        seed : int, optional
            Random seed, by default None.

        Returns
        -------
        A dict of time, event, entry and args (covariates)

        """
        return self._stochastic_process.generate_failure_data(
            nb_samples, time_window, seed
        )

    def sample(self, nb_samples: int, time_window: tuple[float, float], seed=None):
        """Renewal data sampling.

        This function will sample data and encapsulate them in an object.

        Parameters
        ----------
        nb_samples : int
            The number of samples.
        time_window : tuple of two floats
            Time window in which data are sampled.
        seed : int, optional
            Random seed, by default None.

        """

        return self._stochastic_process.sample(nb_samples, time_window, seed)


class NonHomogeneousPoissonAgeReplacementPolicy(ReplacementPolicy):
    r"""Age replacement policy for non-Homogeneous Poisson process.

    Parameters
    ----------
    nhpp : non-homogeneous Poisson process
        The underlying non homogeneous poisson process. If the process expects covars, it must be frozen before.
    cr : float or 1darray
        The cost of repair.
    cp : float or 1darray
        The cost of failure.
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    ar : float or 1darray, optional
        Ages of preventive replacements, by default None. If not given, one must call ``optimize`` to set ``ar`` values
        and access to the rest of the object interface.

    Attributes
    ----------
    cp
    cr
    ar
    """

    def __init__(self, nhpp, cr, cp, discounting_rate=0.0, ar=None):
        super().__init__(
            nhpp,
            cost_structure={"cr": reshape_1d_arg(cr), "cp": reshape_1d_arg(cp)},
            discounting_rate=discounting_rate,
        )
        self.ar = ar

    @property
    def cp(self):
        """Costs of preventive replacement.

        Returns
        -------
        np.ndarray
        """
        return flatten_if_possible(self._cost_structure["cp"])

    @cp.setter
    def cp(self, value):
        self._cost_structure["cp"] = reshape_1d_arg(value)

    @property
    def cr(self):
        """Cost of minimal repair.

        Returns
        -------
        np.ndarray
        """
        return flatten_if_possible(self._cost_structure["cr"])

    @cr.setter
    def cr(self, value):
        self._cost_structure["cr"] = reshape_1d_arg(value)

    @property
    def ar(self):
        """Preventive ages of replacement.

        Returns
        -------
        np.ndarray
        """
        if self._ar is None:
            return self._ar
        return flatten_if_possible(self._ar)

    @ar.setter
    def ar(self, value):
        if value is not None:
            value = reshape_1d_arg(value)
            self._ar = value
        else:
            self._ar = None

    def expected_net_present_value(self, tf, nb_steps, total_sum=False):
        raise NotImplemented("implementation will come in a future release")

    def asymptotic_expected_net_present_value(self, total_sum=False):
        raise NotImplemented("implementation will come in a future release")

    def expected_equivalent_annual_cost(self, tf, nb_steps, total_sum=False):
        raise NotImplemented("implementation will come in a future release")

    def asymptotic_expected_equivalent_annual_cost(self):
        discounting = ExponentialDiscounting(self.discounting_rate)

        if self.ar is None:
            raise ValueError

        if self.discounting_rate == 0.0:
            asymptotic_eeac = (
                self.cp
                + self.cr
                * legendre_quadrature(
                    lambda t: self.baseline_model.intensity(t), 0, self._ar
                )
            ) / self._ar
        else:
            asymptotic_eeac = (
                self.discounting_rate
                * (
                    self.cp * discounting.factor(self._ar)
                    + self.cr
                    * legendre_quadrature(
                        lambda t: discounting.factor(t)
                        * self.baseline_model.intensity(t),
                        0,
                        self._ar,
                    )
                )
                / (1 - discounting.factor(self._ar))
            )
        return np.squeeze(asymptotic_eeac)  # () or (m,)

    def optimize(self):
        """
        Optimize the policy according the costs, the discounting rate and the underlying non-homogeneous Poisson process.

        Returns
        -------
        Self
             Optimized policy.
        """

        discounting = ExponentialDiscounting(self.discounting_rate)
        x0 = np.atleast_2d(self.baseline_model.lifetime_model.mean())

        def eq(a):
            if discounting.rate != 0:
                return (
                    (1 - discounting.factor(a))
                    / self.discounting_rate
                    * self.baseline_model.intensity(a)
                    - legendre_quadrature(
                        lambda t: discounting.factor(t)
                        * self.baseline_model.intensity(t),
                        np.array(0.0),
                        a,
                    )
                    - self._cost_structure["cp"] / self._cost_structure["cr"]
                )
            return (
                a * self.baseline_model.intensity(a)
                - self.baseline_model.cumulative_intensity(a)
                - self._cost_structure["cp"] / self._cost_structure["cr"]
            )

        self.ar = newton(eq, x0)
        return self



# pyright: basic

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable

import numpy as np
from numpy.typing import NDArray
from typing_extensions import override

from relife.utils import get_model_nb_assets

from ._iterators import (
    Kijima1ProcessIterator,
    Kijima2ProcessIterator,
    NonHomogeneousPoissonProcessIterator,
    RenewalProcessIterator,
    RenewalRewardProcessIterator,
    StochasticDataIterator,
)

__all__ = [
    "StochasticDataIterable",
    "RenewalProcessIterable",
    "NonHomogeneousPoissonProcessIterable",
    "Kijima1ProcessIterable",
    "Kijima2ProcessIterable",
]


class StochasticDataIterable(Iterable[NDArray[np.void]], ABC):
    def __init__(
        self,
        nb_samples: int,
        time_window: tuple[float, float],
        seed: int | None = None,
    ):
        t0, tf = time_window
        if t0 < 0 or tf < 0 or t0 > tf:
            raise ValueError(
                f"Incorrect time window. Got {time_window}. Values must be positive and first value can't lower than second value."
            )
        self.time_window = t0, tf
        self.nb_samples = nb_samples
        self.seed = seed

    @property
    def t0(self) -> float:
        return self.time_window[0]

    @property
    def tf(self) -> float:
        return self.time_window[1]

    @override
    @abstractmethod
    def __iter__(self) -> StochasticDataIterator: ...


class RenewalProcessIterable(StochasticDataIterable):
    def __init__(
        self,
        process,
        nb_samples: int,
        time_window: tuple[float, float],
        seed: int | None = None,
    ):
        super().__init__(nb_samples, time_window, seed=seed)
        self.process = process

    def __iter__(self) -> RenewalProcessIterator:
        from relife.stochastic_process import RenewalProcess, RenewalRewardProcess

        if isinstance(self.process, RenewalProcess):
            return RenewalProcessIterator(
                self.process,
                self.nb_samples,
                self.time_window,
                nb_assets=get_model_nb_assets(self.process),
                seed=self.seed,
            )
        if isinstance(self.process, RenewalRewardProcess):
            return RenewalRewardProcessIterator(
                self.process,
                self.nb_samples,
                self.time_window,
                nb_assets=get_model_nb_assets(self.process),
                seed=self.seed,
            )
        raise ValueError


class NonHomogeneousPoissonProcessIterable(StochasticDataIterable):
    def __init__(
        self,
        process,
        nb_samples: int,
        time_window: tuple[float, float],
        seed: int | None = None,
    ):
        super().__init__(nb_samples, time_window, seed=seed)
        self.process = process

    def __iter__(self) -> NonHomogeneousPoissonProcessIterator:
        return NonHomogeneousPoissonProcessIterator(
            self.process,
            self.nb_samples,
            self.time_window,
            nb_assets=get_model_nb_assets(self.process),
            seed=self.seed,
        )


class Kijima1ProcessIterable(StochasticDataIterable):
    def __init__(
        self,
        process,
        nb_samples: int,
        time_window: tuple[float, float],
        seed: int | None = None,
    ):
        super().__init__(nb_samples, time_window, seed=seed)
        self.process = process

    def __iter__(self) -> Kijima1ProcessIterator:
        return Kijima1ProcessIterator(
            self.process,
            self.nb_samples,
            self.time_window,
            nb_assets=get_model_nb_assets(self.process),
            seed=self.seed,
        )


class Kijima2ProcessIterable(StochasticDataIterable):
    def __init__(
        self,
        process,
        nb_samples: int,
        time_window: tuple[float, float],
        seed: int | None = None,
    ):
        super().__init__(nb_samples, time_window, seed=seed)
        self.process = process

    def __iter__(self) -> Kijima2ProcessIterator:
        return Kijima2ProcessIterator(
            self.process,
            self.nb_samples,
            self.time_window,
            nb_assets=get_model_nb_assets(self.process),
            seed=self.seed,
        )



# pyright: basic
from ._array_utils import (
    flatten_if_possible,
    get_args_nb_assets,
    reshape_1d_arg,
)
from ._model_checks import get_model_nb_assets

__all__ = [
    "get_args_nb_assets",
    "reshape_1d_arg",
    "flatten_if_possible",
    "get_model_nb_assets",
]


# pyright: basic
from __future__ import annotations

from abc import ABC
from typing import Any, Literal, Optional, overload

import numpy as np
from numpy.typing import NDArray

from relife.economic import RunToFailureReward
from relife.lifetime_model import (
    LeftTruncatedModel,
)
from relife.stochastic_process import RenewalRewardProcess
from relife.typing import (
    AnyFloat,
    AnyParametricLifetimeModel,
    NumpyFloat,
)
from relife.utils import flatten_if_possible, reshape_1d_arg

from ._base import OneCycleExpectedCosts, ReplacementPolicy


@overload
def run_to_failure_policy(
    baseline_model: AnyParametricLifetimeModel[()],
    cf: AnyFloat,
    one_cycle: Literal[True],
    **kwargs: Any,
) -> OneCycleRunToFailurePolicy: ...
@overload
def run_to_failure_policy(
    baseline_model: AnyParametricLifetimeModel[()],
    cf: AnyFloat,
    one_cycle: Literal[False],
    **kwargs: Any,
) -> RunToFailurePolicy: ...
@overload
def run_to_failure_policy(
    baseline_model: AnyParametricLifetimeModel[()],
    cf: AnyFloat,
    one_cycle: bool = False,
    **kwargs: Any,
) -> OneCycleRunToFailurePolicy | RunToFailurePolicy: ...
def run_to_failure_policy(
    baseline_model: AnyParametricLifetimeModel[()],
    cf: AnyFloat,
    one_cycle: bool = False,
    **kwargs: Any,
) -> OneCycleRunToFailurePolicy | RunToFailurePolicy:
    """
    Creates a run-to-failure policy.

    Parameters
    ----------
    baseline_model : parametric model
        Parametric model required by the policy.
    cf : float or 1d-array
        Cost of failure.
    one_cycle : bool, default False
        If True, returns the one cycle variation of the policy.
    **kwargs
        Extra arguments required by the policy (a0, discounting_rate, etc.)

    Returns
    -------
    Policy
        Policy corresponding to the ``baseline_model`` and the ``cost_structure``.

    Raises
    ------
    ValueError
        If ``baseline_model`` or ``cost_structure`` does not have a corresponding policy.
    """
    if one_cycle:
        return OneCycleRunToFailurePolicy(baseline_model, cf, **kwargs)
    return RunToFailurePolicy(baseline_model, cf, **kwargs)


class BaseRunToFailure(ReplacementPolicy[AnyParametricLifetimeModel[()]], ABC):
    _cost_structure: dict[str, NumpyFloat]
    _a0: Optional[NumpyFloat]
    discounting_rate: float

    def __init__(
        self,
        lifetime_model: AnyParametricLifetimeModel[()],
        cf: AnyFloat,
        discounting_rate: float = 0.0,
        a0: Optional[AnyFloat] = None,
    ):
        super().__init__(
            lifetime_model,
            {"cf": reshape_1d_arg(cf)},
            discounting_rate=discounting_rate,
        )
        self._a0 = reshape_1d_arg(a0) if a0 is not None else a0

    @property
    def a0(self) -> Optional[NumpyFloat]:
        """Current ages of the assets.

        Returns
        -------
        np.ndarray
        """
        # _a0 is (m, 1) but exposed cf is (m,)
        if self._a0 is None:
            return self._a0
        return flatten_if_possible(self._a0)

    @property
    def cf(self) -> NumpyFloat:
        """Cost of failure.

        Returns
        -------
        np.ndarray
        """
        # _cf is (m, 1) but exposed cf is (m,)
        return flatten_if_possible(self._cost_structure["cf"])

    @cf.setter
    def cf(self, value: AnyFloat) -> None:
        self._cost_structure["cf"] = reshape_1d_arg(value)


class OneCycleRunToFailurePolicy(BaseRunToFailure):
    r"""One cyle run-to-failure policy.

    Asset is replaced upon failure with cost :math:`c_f`.

    .. note::

        ``OneCycleRunToFailurePolicy`` differs from ``RunToFailurePolicy``
        because only one cycle of replacement is considered.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    cf : float or 1darray
        Costs of failures
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    a0 : float or 1darray, optional
        Current ages of the assets, by default 0 for each asset. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.

    Attributes
    ----------
    cf
    """

    def __init__(
        self,
        lifetime_model: AnyParametricLifetimeModel[()],
        cf: AnyFloat,
        discounting_rate: float = 0.0,
        a0: Optional[AnyFloat] = None,
        period_before_discounting: float = 1.0,
    ) -> None:
        super().__init__(lifetime_model, cf, discounting_rate=discounting_rate, a0=a0)
        self.period_before_discounting = period_before_discounting

    @property
    def _expected_costs(self) -> OneCycleExpectedCosts:
        if self.a0 is None:
            return OneCycleExpectedCosts(
                self.baseline_model,
                RunToFailureReward(self.cf),
                discounting_rate=self.discounting_rate,
                period_before_discounting=self.period_before_discounting,
            )
        return OneCycleExpectedCosts(
            LeftTruncatedModel(self.baseline_model).freeze(self.a0),
            RunToFailureReward(self.cf),
            discounting_rate=self.discounting_rate,
            period_before_discounting=self.period_before_discounting,
        )

    def expected_net_present_value(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        timeline, npv = self._expected_costs.expected_net_present_value(tf, nb_steps)
        if total_sum and npv.ndim == 2:
            return timeline, np.sum(npv, axis=0)
        return timeline, npv

    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        asymptotic_npv = self._expected_costs.asymptotic_expected_net_present_value()
        if total_sum:
            return np.sum(asymptotic_npv)
        return asymptotic_npv

    def expected_equivalent_annual_cost(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        timeline, eeac = self._expected_costs.expected_equivalent_annual_cost(
            tf, nb_steps
        )
        if total_sum and eeac.ndim == 2:
            return timeline, np.sum(eeac, axis=0)
        return timeline, eeac

    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        asymptotic_eeac = (
            self._expected_costs.asymptotic_expected_equivalent_annual_cost()
        )
        if total_sum:
            return np.sum(asymptotic_eeac)
        return asymptotic_eeac


class RunToFailurePolicy(BaseRunToFailure):
    r"""Run-to-failure renewal policy.

    Asset is replaced upon failure with cost :math:`c_f`.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    cf : float or 1darray
        Costs of failures
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    a0 : float or 1darray, optional
        Current ages of the assets, by default 0 for each asset. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.

    Attributes
    ----------
    cf

    References
    ----------
    .. [1] Van der Weide, J. A. M., & Van Noortwijk, J. M. (2008). Renewal
        theory with exponential and hyperbolic discounting. Probability in
        the Engineering and Informational Sciences, 22(1), 53-74.
    """

    @property
    def _stochastic_process(self) -> RenewalRewardProcess:
        if self.a0 is None:
            return RenewalRewardProcess(
                self.baseline_model,
                RunToFailureReward(self.cf),
                discounting_rate=self.discounting_rate,
            )
        return RenewalRewardProcess(
            self.baseline_model,
            RunToFailureReward(self.cf),
            discounting_rate=self.discounting_rate,
            first_lifetime_model=LeftTruncatedModel(self.baseline_model).freeze(
                self.a0
            ),
        )

    def expected_net_present_value(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        timeline, npv = self._stochastic_process.expected_total_reward(tf, nb_steps)
        if total_sum and npv.ndim == 2:
            npv = np.sum(npv, axis=0)
        return timeline, npv

    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_net_present_value(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        asymptotic_npv = self._stochastic_process.asymptotic_expected_total_reward()
        if total_sum:
            return np.sum(asymptotic_npv)
        return asymptotic_npv

    def expected_equivalent_annual_cost(
        self, tf: float, nb_steps: int, total_sum: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        timeline, eeac = self._stochastic_process.expected_equivalent_annual_worth(
            tf, nb_steps
        )
        if total_sum and eeac.ndim == 2:
            eeac = np.sum(eeac, axis=0)
        return timeline, eeac

    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[False]
    ) -> NDArray[np.float64]: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: Literal[True]
    ) -> np.float64: ...
    @overload
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]: ...
    def asymptotic_expected_equivalent_annual_cost(
        self, total_sum: bool = False
    ) -> np.float64 | NDArray[np.float64]:
        asymptotic_eeac = (
            self._stochastic_process.asymptotic_expected_equivalent_annual_worth()
        )
        if total_sum:
            return np.sum(asymptotic_eeac)
        return asymptotic_eeac

    def sample(self, size, tf, t0=0.0, seed=None):
        """Renewal data sampling.

        This function will sample data and encapsulate them in an object.

        Parameters
        ----------
        size : int
            The size of the desired sample.
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        size : int or tuple of 2 int
            Size of the sample
        seed : int, optional
            Random seed, by default None.

        """
        return self._stochastic_process.sample(tf, t0, size, seed)



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("relife")
except PackageNotFoundError:
    # package is not installed
    pass

__all__: list[str] = []

_submodules = [
    "data",
    "lifetime_model",
    "likelihood",
    "policy",
    "stochastic_process",
    "quadrature",
    "economic",
    "typing",
    "utils",
]

__all__ += _submodules



from __future__ import annotations

from typing import Callable, Literal, TypeVarTuple, overload

import numpy as np
import numpydoc.docscrape as docscrape  # pyright: ignore[reportMissingTypeStubs]
from numpy.typing import NDArray
from typing_extensions import override

from relife.base import is_frozen
from relife.typing import (
    AnyFloat,
    AnyParametricLifetimeModel,
    NumpyBool,
    NumpyFloat,
    Seed,
)
from relife.utils import reshape_1d_arg

from ._base import FrozenParametricLifetimeModel, ParametricLifetimeModel, document_args

__all__: list[str] = ["AgeReplacementModel", "LeftTruncatedModel"]

Ts = TypeVarTuple("Ts")

_ar_args_docstring = [
    docscrape.Parameter(
        "ar",
        "float or np.ndarray",
        [
            "Age of replacement values.",
            """
            If ndarray, shape can only be `(m,)` because only one age of
            replacement per asset can be given.
            """,
        ],
    ),
    docscrape.Parameter(
        "*args",
        "",
        [
            "Any other arguments needed by the model.",
        ],
    ),
]


class AgeReplacementModel(ParametricLifetimeModel[*tuple[AnyFloat, *Ts]]):
    r"""
    Age replacement model.

    Lifetime model where the assets are replaced at age :math:`a_r`. This is
    equivalent to the model of :math:`\min(X,a_r)` where :math:`X` is a
    baseline lifetime and :math:`a_r` is the age of replacement.

    Parameters
    ----------
    baseline : any parametric lifetime model (frozen lifetime model works)
        The base lifetime model without conditional probabilities

    Attributes
    ----------
    baseline
    nb_params
    params
    params_names
    plot
    """

    baseline: AnyParametricLifetimeModel[*Ts]

    def __init__(self, baseline: AnyParametricLifetimeModel[*Ts]):
        super().__init__()
        self.baseline = baseline

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def sf(self, time: AnyFloat, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return np.where(time < ar, self.baseline.sf(time, *args), 0.0)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def hf(self, time: AnyFloat, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return np.where(time < ar, self.baseline.hf(time, *args), 0.0)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def cdf(self, time: AnyFloat, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return super().cdf(time, *(ar, *args))

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def chf(self, time: AnyFloat, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return np.where(time < ar, self.baseline.chf(time, *args), 0.0)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def isf(self, probability: AnyFloat, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return np.minimum(self.baseline.isf(probability, *args), ar)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def ichf(
        self, cumulative_hazard_rate: AnyFloat, ar: AnyFloat, *args: *Ts
    ) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return np.minimum(self.baseline.ichf(cumulative_hazard_rate, *args), ar)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def pdf(self, time: AnyFloat, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return np.where(time < ar, self.baseline.pdf(time, *args), 0)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def mrl(self, time: AnyFloat, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        ub = np.array(np.inf)
        # ar.shape == (m, 1)
        mask = time >= ar  # (m, 1) or (m, n)
        if np.any(mask):
            time, ub = np.broadcast_arrays(time, ub)
            time = np.ma.MaskedArray(time, mask)  # (m, 1) or (m, n)
            ub = np.ma.MaskedArray(ub, mask)  # (m, 1) or (m, n)
        mu = self.ls_integrate(
            lambda x: x - time, time, ub, ar, *args, deg=10
        ) / self.sf(time, ar, *args)  # () or (n,) or (m, n)
        np.ma.filled(mu, 0)
        return np.ma.getdata(mu)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def ppf(self, probability: AnyFloat, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return self.isf(1 - probability, ar, *args)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def median(self, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return self.ppf(np.array(0.5), ar, *args)

    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        ar: AnyFloat,
        *args: *Ts,
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> NumpyFloat: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        ar: AnyFloat,
        *args: *Ts,
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        ar: AnyFloat,
        *args: *Ts,
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        ar: AnyFloat,
        *args: *Ts,
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        ar: AnyFloat,
        *args: *Ts,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ): ...
    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def rvs(
        self,
        size: int | tuple[int, int],
        ar: AnyFloat,
        *args: *Ts,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ):
        ar = reshape_1d_arg(ar)
        baseline_rvs = self.baseline.rvs(
            size,
            *args,
            return_event=return_event,
            return_entry=return_entry,
            seed=seed,
        )
        time = baseline_rvs[0] if isinstance(baseline_rvs, tuple) else baseline_rvs
        time = np.minimum(time, ar)  # it may change time shape by broadcasting
        if not return_event and not return_entry:
            return time
        elif return_event and not return_entry:
            event = np.broadcast_to(baseline_rvs[1], time.shape).copy()
            event = np.where(time != ar, event, ~event)
            return time, event
        elif not return_event and return_entry:
            entry = np.broadcast_to(baseline_rvs[1], time.shape).copy()
            return time, entry
        else:
            event, entry = baseline_rvs[1:]
            event = np.broadcast_to(event, time.shape).copy()
            entry = np.broadcast_to(entry, time.shape).copy()
            event = np.where(time != ar, event, ~event)
            return time, event, entry

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def ls_integrate(
        self,
        func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
        a: AnyFloat,
        b: AnyFloat,
        ar: AnyFloat,
        *args: *Ts,
        deg: int = 10,
    ) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        b = np.minimum(ar, b)
        integration = self.baseline.ls_integrate(func, a, b, *args, deg=deg)
        if func(ar).ndim == 2 and integration.ndim == 1:
            integration = integration.reshape(-1, 1)
        return integration + np.where(
            b == ar, func(ar) * self.baseline.sf(ar, *args), 0
        )

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def moment(self, n: int, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return self.ls_integrate(
            lambda x: x**n,
            np.float64(0),
            np.inf,
            ar,
            *args,
            deg=100,
        )

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def mean(self, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return self.moment(1, ar, *args)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_ar_args_docstring)
    def var(self, ar: AnyFloat, *args: *Ts) -> NumpyFloat:
        ar = reshape_1d_arg(ar)
        return self.moment(2, ar, *args) - self.moment(1, ar, *args) ** 2

    def freeze(
        self, ar: AnyFloat, *args: *Ts
    ) -> FrozenParametricLifetimeModel[*tuple[AnyFloat, *Ts]]:
        """
        Freeze age replacement values and other arguments into the object data.

        Parameters
        ----------
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        FrozenParametricModel
        """
        return FrozenParametricLifetimeModel(self, ar, *args)


_a0_args_docstring = [
    docscrape.Parameter(
        "a0",
        "float or np.ndarray",
        [
            "Current ages.",
            "If ndarray, shape can only be `(m,)`.",
        ],
    ),
    docscrape.Parameter(
        "*args",
        "",
        [
            "Any other arguments needed by the model.",
        ],
    ),
]


class LeftTruncatedModel(ParametricLifetimeModel[*tuple[AnyFloat, *Ts]]):
    r"""Left truncated model.

    Lifetime model where the assets have already reached the age :math:`a_0`.

    Parameters
    ----------
    baseline : any parametric lifetime model (frozen lifetime model works)
        The base lifetime model without conditional probabilities
    nb_params
    params
    params_names
    plot

    Attributes
    ----------
    baseline
    nb_params
    params
    params_names
    plot
    """

    baseline: AnyParametricLifetimeModel[*Ts]

    def __init__(self, baseline: AnyParametricLifetimeModel[*Ts]):
        super().__init__()
        self.baseline = baseline

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def sf(self, time: AnyFloat, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().sf(time, a0, *args)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def pdf(self, time: AnyFloat, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().pdf(time, a0, *args)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def isf(self, probability: AnyFloat, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        cumulative_hazard_rate = -np.log(probability + 1e-6)  # avoid division by zero
        a0 = reshape_1d_arg(a0)
        return self.ichf(cumulative_hazard_rate, a0, *args)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def chf(self, time: AnyFloat, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return self.baseline.chf(a0 + time, *args) - self.baseline.chf(a0, *args)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def cdf(self, time: AnyFloat, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().cdf(time, *(a0, *args))

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def hf(self, time: AnyFloat, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return self.baseline.hf(a0 + time, *args)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def ichf(
        self, cumulative_hazard_rate: AnyFloat, a0: AnyFloat, *args: *Ts
    ) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return (
            self.baseline.ichf(
                cumulative_hazard_rate + self.baseline.chf(a0, *args), *args
            )
            - a0
        )

    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        a0: AnyFloat,
        *args: *Ts,
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> NumpyFloat: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        a0: AnyFloat,
        *args: *Ts,
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        a0: AnyFloat,
        *args: *Ts,
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        a0: AnyFloat,
        *args: *Ts,
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        a0: AnyFloat,
        *args: *Ts,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ): ...
    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def rvs(
        self,
        size: int | tuple[int, int],
        a0: AnyFloat,
        *args: *Ts,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ):
        a0 = reshape_1d_arg(a0)
        super_rvs = super().rvs(
            size,
            *(a0, *args),
            return_event=return_event,
            return_entry=return_entry,
            seed=seed,
        )
        time = super_rvs[0] if isinstance(super_rvs, tuple) else super_rvs
        complete_ages = time + a0
        output = [
            time,
        ]  # at least time in output
        if return_event:
            event = super_rvs[1]  # event always at index 1
            # reconstruct event for AgeReplacementModel c omposition as super skips this info
            if isinstance(self.baseline, AgeReplacementModel):
                ar = reshape_1d_arg(args[0])
                event = np.where(complete_ages < ar, event, ~event)
            if is_frozen(self.baseline) and isinstance(
                self.baseline.unfreeze(), AgeReplacementModel
            ):
                ar = reshape_1d_arg(self.baseline.args[0])
                event = np.where(complete_ages < ar, event, ~event)
            output.append(event)
        if return_entry:
            output[0] = complete_ages  # don't return residual ages
            entry = super_rvs[-1]  # entry always at last index
            entry = np.broadcast_to(a0, entry.shape).copy()
            output.append(entry)
        if len(output) > 1:
            return tuple(output)  # return tuple, not list
        return output[0]

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def ls_integrate(
        self,
        func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
        a: AnyFloat,
        b: AnyFloat,
        a0: AnyFloat,
        *args: *Ts,
        deg: int = 10,
    ) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().ls_integrate(func, a, b, *(a0, *args), deg=deg)

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def mean(self, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().mean(*(a0, *args))

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def median(self, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().median(*(a0, *args))

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def var(self, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().var(*(a0, *args))

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def moment(self, n: int, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().moment(n, *(a0, *args))

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def mrl(self, time: AnyFloat, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().mrl(time, *(a0, *args))

    @override
    @document_args(base_cls=ParametricLifetimeModel, args_docstring=_a0_args_docstring)
    def ppf(self, probability: AnyFloat, a0: AnyFloat, *args: *Ts) -> NumpyFloat:
        a0 = reshape_1d_arg(a0)
        return super().ppf(probability, *(a0, *args))

    def freeze(
        self, a0: AnyFloat, *args: *Ts
    ) -> FrozenParametricLifetimeModel[*tuple[AnyFloat, *Ts]]:
        """
        Freeze conditional age values and other arguments into the object data.

        Parameters
        ----------
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If
            ndarray, shape can only be (m,) as only one age per asset can be
            given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        FrozenLeftTruncatedModel
        """
        return FrozenParametricLifetimeModel(self, a0, *args)



from .kijima_process import Kijima1Process, Kijima2Process
from .non_homogeneous_poisson_process import (
    NonHomogeneousPoissonProcess,
)
from .renewal_process import RenewalProcess, RenewalRewardProcess

__all__ = [
    "RenewalProcess",
    "RenewalRewardProcess",
    "NonHomogeneousPoissonProcess",
    "Kijima1Process",
    "Kijima2Process",
]



"""
``relife.typing``
=================

The ReLife typing module exposes various types to type hint codes
using relife.

"""

from ._models import AnyParametricLifetimeModel
from ._random import Seed
from ._scalars import AnyFloat, NumpyBool, NumpyFloat

__all__ = [
    "AnyParametricLifetimeModel",
    "Seed",
    "AnyFloat",
    "NumpyBool",
    "NumpyFloat",
]



# pyright: basic

import numpy as np

from relife.stochastic_process import Kijima1Process, Kijima2Process
from relife.stochastic_process._sample._iterables import (
    Kijima1ProcessIterable,
    Kijima2ProcessIterable,
)
from relife.stochastic_process._sample.tests.utils import select_from_struct


class TestBasicSampling:
    def test_kijima_1(self, distribution):
        kijima_1 = Kijima1Process(distribution, q=0.5)
        nb_samples = 10
        t0 = distribution.ppf(0.3)
        tf = distribution.ppf(0.95)

        iterable = Kijima1ProcessIterable(kijima_1, nb_samples, (t0, tf))
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(
            struct_array, order=("asset_id", "sample_id", "timeline")
        )

        # Check Kijima I property for each sample
        for i in range(nb_samples):
            select_sample = select_from_struct(struct_array, sample_id=i)
            np.testing.assert_almost_equal(
                kijima_1.q * select_sample["timeline"][:-1], select_sample["entry"][1:]
            )

    def test_kijima_2(self, distribution):
        kijima_2 = Kijima2Process(distribution, q=0.5)
        nb_samples = 10
        t0 = distribution.ppf(0.3)
        tf = distribution.ppf(0.95)

        iterable = Kijima2ProcessIterable(kijima_2, nb_samples, (t0, tf))
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(
            struct_array, order=("asset_id", "sample_id", "timeline")
        )

        # Check Kijima II property for each sample
        for i in range(nb_samples):
            select_sample = select_from_struct(struct_array, sample_id=i)
            np.testing.assert_almost_equal(
                kijima_2.q * select_sample["time"][:-1], select_sample["entry"][1:]
            )



# pyright: basic

from relife.economic import AgeReplacementReward, RunToFailureReward


def test_run_to_failure_reward(time, cf, expected_out_shape):
    reward = RunToFailureReward(cf)
    assert reward.conditional_expectation(time).shape == expected_out_shape(
        time=time, cf=cf
    )
    assert reward.sample(time).shape == expected_out_shape(time=time, cf=cf)


def test_age_replacement_reward(time, cf, cp, ar, expected_out_shape):
    reward = AgeReplacementReward(cf, cp, ar)
    assert reward.conditional_expectation(time).shape == expected_out_shape(
        time=time, cf=cf, cp=cp, ar=ar
    )
    assert reward.sample(time).shape == expected_out_shape(
        time=time, cf=cf, cp=cp, ar=ar
    )



# pyright: basic

from abc import ABC, abstractmethod

import numpy as np
from numpy.typing import NDArray

from relife.typing import AnyFloat
from relife.utils import reshape_1d_arg


class Reward(ABC):
    @abstractmethod
    def conditional_expectation(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        """Conditional expected reward"""
        pass

    @abstractmethod
    def sample(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        """Reward conditional sampling.

        Parameters
        ----------
        time : ndarray
            Time duration values.


        Returns
        -------
        ndarray
            Random drawing of a reward with respect to time.
        """


class RunToFailureReward(Reward):
    r"""Run-to-failure reward.

    Parameters
    ----------
    cf : float or 1darray
        The cost of failure.

    Attributes
    ----------
    cf
    """

    cf: np.float64 | NDArray[np.float64]

    def __init__(self, cf: AnyFloat) -> None:
        self.cf = reshape_1d_arg(cf)

    def conditional_expectation(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.ones_like(time) * self.cf

    def sample(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        return self.conditional_expectation(time)


class AgeReplacementReward(Reward):
    r"""Age replacement reward.

    Parameters
    ----------
    cf : float or 1darray
        The cost of failure.
    cp : float or 1darray
        The cost of preventive replacement.

    Attributes
    ----------
    cf
    cp
    ar
    """

    cf: np.float64 | NDArray[np.float64]
    cp: np.float64 | NDArray[np.float64]
    ar: np.float64 | NDArray[np.float64]

    def __init__(self, cf: AnyFloat, cp: AnyFloat, ar: AnyFloat) -> None:
        self.cf = reshape_1d_arg(cf)
        self.cp = reshape_1d_arg(cp)
        self.ar = reshape_1d_arg(ar)

    def conditional_expectation(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.where(time < self.ar, self.cf, self.cp)

    def sample(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        return self.conditional_expectation(time)


class Discounting(ABC):
    @abstractmethod
    def factor(self, timeline: NDArray[np.float64]) -> NDArray[np.float64]: ...

    @abstractmethod
    def annuity_factor(self, timeline: NDArray[np.float64]) -> NDArray[np.float64]: ...


class ExponentialDiscounting(Discounting):
    """
    Exponential discounting.

    Parameters
    ----------
    rate : float
        The discounting rate
    """

    rate: float

    def __init__(self, rate: float = 0.0) -> None:
        self.rate = rate

    def factor(self, timeline: NDArray[np.float64]) -> NDArray[np.float64]:
        if self.rate != 0.0:
            return np.exp(-self.rate * timeline)
        else:
            return np.ones_like(timeline)

    def annuity_factor(self, timeline: NDArray[np.float64]) -> NDArray[np.float64]:
        if self.rate != 0.0:
            return (1 - np.exp(-self.rate * timeline)) / self.rate
        else:
            return timeline



"""Additional types used in the codebase for type checking."""

from typing import (
    Callable,
    Literal,
    Protocol,
    Sequence,
    TypeVarTuple,
    overload,
)

import numpy as np
from numpy.typing import NDArray

from ._random import Seed
from ._scalars import AnyFloat, NumpyBool, NumpyFloat

__all__ = ["AnyParametricLifetimeModel"]

Ts = TypeVarTuple("Ts")


class AnyParametricLifetimeModel(Protocol[*Ts]):
    """
    Structural type for any parametric lifetime model.
    It is particularly needed where an parameter can expect
    ParametricLifetimeModel[*Ts] or FrozenParametricLifetimeModel[*Ts]
    Their interfaces are different (the first expects args but the second
    doesn't). See conditional_model.py.
    """

    def sf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat: ...
    def hf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat: ...
    def chf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat: ...
    def pdf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat: ...
    def cdf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat: ...
    def ppf(self, probability: AnyFloat, *args: *Ts) -> NumpyFloat: ...
    def median(self, *args: *Ts) -> NumpyFloat: ...
    def isf(self, probability: AnyFloat, *args: *Ts) -> NumpyFloat: ...
    def ichf(self, cumulative_hazard_rate: AnyFloat, *args: *Ts) -> NumpyFloat: ...
    @overload
    def rvs(
        self,
        size: int | Sequence[int],
        *args: *Ts,
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> NumpyFloat: ...
    @overload
    def rvs(
        self,
        size: int | Sequence[int],
        *args: *Ts,
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool]: ...
    @overload
    def rvs(
        self,
        size: int | Sequence[int],
        *args: *Ts,
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | Sequence[int],
        *args: *Ts,
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | Sequence[int],
        *args: *Ts,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ): ...
    def rvs(
        self,
        size: int | Sequence[int],
        *args: *Ts,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ): ...
    def ls_integrate(
        self,
        func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
        a: NumpyFloat,
        b: NumpyFloat,
        *args: *Ts,
        deg: int = 10,
    ) -> NumpyFloat: ...
    def moment(self, n: int, *args: *Ts) -> NumpyFloat: ...
    def mean(self, *args: *Ts) -> NumpyFloat: ...
    def var(self, *args: *Ts) -> NumpyFloat: ...
    def mrl(self, time: AnyFloat, *args: *Ts) -> NumpyFloat: ...
